#if !defined(__Base_h__)
#define __Base_h__

#include <iostream>
#include <iomanip>
#include <cassert>
#include <vector>
#include <stdexcept>
#include <cmath>
#include <ctime>
#include <map>
namespace My{
using namespace std;

namespace Base
{

/* Class vector represent a contigours block of
Elements, such pointer to them could be obtained */
template<class T, class Real = T>
class Vector
{
public:
	virtual T operator[](int)const
	{
		throw std::out_of_range("no [] operator");
	}
	virtual T& operator[](int)
	{
		throw std::out_of_range("no [] operator");
	}

	virtual int Size() const = 0;
	virtual T* GetPtr() = 0;
	virtual const T* GetPtr() const = 0;
	virtual ~Vector()
	{
	}
	virtual Real NormC()const
	{
		T const* a = GetPtr();

		Real max = 0;
		Real b;
		for (int i = 0; i < Size(); i++)
			if (max < (b=fabs(*a++)))
				max = b;
		return max;
	}
	virtual Real Norm1()const
	{
		T const* a = GetPtr();

		Real max = 0;
		for (int i = 0; i < Size(); i++)
			max += fabs(*a++);
		return max;
	}
	virtual Real Norm2()const
	{
		T const* a = GetPtr();

		Real max = 0;
		for (int i = 0; i < Size(); i++, a++)
			max += *a * *a;
		return sqrt(max);
	}
	virtual void ZeroVector()
	{
		T* a = GetPtr();
		T* ea = a + Size();
		while(a!=ea)
			*a++ = T(0);
	}
};


/*
Class Matrix represents 2d matrix.
we have indexing operator (i, j)
ZeroMatrix() method
Width()
Height()
*/
template<class T, class Real = T>
class Matrix
{
public:
	virtual int Width() const = 0;
	virtual int Height() const = 0;
	virtual T operator()(int, int) const = 0;
	virtual T&operator()(int, int) = 0;
	virtual ~Matrix()
	{
	}
	virtual void ZeroMatrix()
	{
		for (int i = 1; i <= Height(); i++)
			for (int j = 1; j <= Width(); j++)
				(*this)(i, j) = 0;
	}
	virtual void MulMatVec(Vector<T, Real> &R, const Vector<T, Real> &X) const
	{
		assert(Width()==X.Size());
		assert(Height()==R.Size());
		T* r = R.GetPtr();
		for(int i = 1; i<=Height(); i++)
		{
			T sum = 0;
			const T* x = X.GetPtr();
			for (int j = 1; j<= Width(); j++)
				sum+= (*this)(i, j) * *x++;
			*r++ = sum;
		}
	}

};

/*
Res  = X + Y;
*/
template<class T, class Real>
void VecAdd(Vector<T, Real> &Res, const Vector<T, Real> &X, const Vector<T, Real> &Y)
{
	assert(Res.Size()==X.Size());
	assert(X.Size()==Y.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = *x++ + *y++;
	}
}

/*
Res = X + Y*z;
*/
template<class T, class Real>
void VecAdd(Vector<T, Real> &Res, const Vector<T, Real> &X, const Vector<T, Real> &Y, const T& Z)
{
	assert(Res.Size()==X.Size());
	assert(X.Size()==Y.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = *x++ + Z * *y++;
	}
}

/*
Res = X - Y;
*/
template<class T, class Real>
void VecSub(Vector<T, Real> &Res, const Vector<T, Real> &X, const Vector<T, Real> &Y)
{
	assert(Res.Size()==X.Size());
	assert(X.Size()==Y.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = *x++ - *y++;
	}
}

/*
Res = X*z - Y;
*/
template<class T, class Real>
void VecMulSub(Vector<T, Real> &Res, const Vector<T, Real> &X, const T& Z, const Vector<T, Real> &Y)
{
	assert(Res.Size()==X.Size());
	assert(X.Size()==Y.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = Z * *x++ - *y++;
	}
}

/*
Res = X*Y;
*/
template<class T, class Real>
void VecMul(Vector<T, Real> &Res, const Vector<T, Real> &X, const Vector<T, Real> &Y)
{
	assert(Res.Size()==X.Size());
	assert(X.Size()==Y.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = *x++ * *y++;
	}
}

/*
Res = X*y;
*/
template<class T, class Real>
void VecMul(Vector<T, Real> &Res, const Vector<T, Real> &X, const T& Y)
{
	assert(Res.Size()==X.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = *x++ * Y;
	}
}

/*
Res = X+y;
*/
template<class T, class Real>
void VecAdd(Vector<T, Real> &Res, const Vector<T, Real> &X, const T& Y)
{
	assert(Res.Size()==X.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = *x++ + Y;
	}

}

/*
Res = y - X;
*/
template<class T, class Real>
void VecSub(Vector<T, Real> &Res, const T& Y, const Vector<T, Real> &X)
{
	assert(Res.Size()==X.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ = Y - *x++;
	}

}

/*
Res += X*Y;
*/
template<class T, class Real>
void VecMulAdd(Vector<T, Real> &Res, const Vector<T, Real> &X, const Vector<T, Real> &Y)
{
	assert(Res.Size()==X.Size());
	assert(X.Size()==Y.Size());

	T* res = Res.GetPtr();
	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	T* eres = res + X.Size();

	while(res<eres)
	{
		*res++ += *x++ * *y++;
	}

}

/*
  returns X.Y
*/
template<class T, class Real>
T VecDot(const Vector<T, Real> &X, const Vector<T, Real> &Y)
{
	assert(X.Size()==Y.Size());

	const T* x = X.GetPtr();
	const T* y = Y.GetPtr();
	const T* ex = x + X.Size();

	T res = 0;
	while(x<ex)
	{
		res += *x++ * *y++;
	}
	return res;
}


template<class T, class Real = T>
class __CSWVector :public Vector<T, Real>
{
	const T*buf;
	const int size;
	__CSWVector();
	T* GetPtr()
	{
		return NULL;
	}
public:
	__CSWVector(const T* where, int sz):buf(where), size(sz)
	{
		assert(sz>0);
		assert(where!=NULL);
	}
	~__CSWVector()
	{
	}
	int Size()const
	{
		return size;
	}
	const T* GetPtr()const
	{
		return buf;
	}
};

/*
Represents a Vector, without own memory
Consturcts from 
(buf, size)
*/
template<class T, class Real = T>
class SWVector :public Vector<T, Real>
{
	T*buf;
	const int size;
	SWVector();
public:
	typedef __CSWVector<T, Real> const constant;
	SWVector(T* where, int sz):buf(where), size(sz)
	{
		assert(sz>0);
		assert(where!=NULL);
	}
	~SWVector()
	{
	}
	int Size()const
	{
		return size;
	}
	T* GetPtr()
	{
		return buf;
	}
	const T* GetPtr()const
	{
		return buf;
	}
};

/*
Represents a full vector.
Memory is own managed
No copy constructors & operators
*/
template<class T, class Real = T>
class FullVector :public Vector<T, Real>
{
	T*buf;
	int size;
	FullVector();
public:
	typedef FullVector<T, Real> VectorType;
	FullVector(const FullVector<T, Real>&a, bool copy = true)
	{
		size = a.size;
		buf = new T[size];
		if (copy)
			for (int i = 0; i < size; i++)
				buf[i] = a.buf[i];
	}
	FullVector(int sz):size(sz)
	{
		assert(sz>0);
		buf = new T[size];		
	}
	void SetSize(int sz)
	{
		assert(sz>0);
		delete[] buf;
		size = sz;
		buf = new T[size];
	}
	~FullVector()
	{
		delete[] buf;
	}
	int Size()const
	{
		return size;
	}
	T* GetPtr()
	{
		return buf;
	}
	T& operator[](int i)
	{
		assert((i>0)&&(i<=size));
		return buf[i-1];
	}
	T operator[](int i)const
	{
		assert((i>0)&&(i<=size));
		return buf[i-1];
	}
	const T* GetPtr()const
	{
		return buf;
	}
};


/* Represents a matrix
With nonzero elemnts only on few diagonals
*/
template<class T, class Real = T>
class DiagMatrix :public Matrix<T, Real>
{
protected:
	int N;
        std::map<int, T*> diags;
	mutable T holder; // here  iz stored;
public:
	const map<int, T*>& GetDiags() const
	{
		return diags;
	}
	DiagMatrix(const vector<int> &v, int n):N(n)
	{
		for (int i = 0; i < v.size(); i++)
		{
			diags[v[i]] = new T[n-abs(v[i])];
		}
	}
	DiagMatrix(int n):N(n)
	{
	}
	
	~DiagMatrix()
	{
	  typename std::map<int, T*>::iterator it;
		for (it = diags.begin(); it!=diags.end(); it++)
		{
			delete[] it->second;
		}
	}
	int Width() const
	{
		return N;
	}
	int Height() const 
	{
		return N;
	}
	void NewMtx(int n)
	{
	  typename std::map<int, T*>::iterator it;
		for (it = diags.begin(); it!=diags.end(); it++)
		{
			delete[] it->second;
		}
		diags.clear();
		N = n;
	}
	virtual void AddDiag(int diag)
	{
		diags[diag] = new T[N-abs(diag)];
	}
	T operator()(int i, int j) const 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=N);
		assert(j<=N);
		typename map<int, T*>::const_iterator it = diags.find(j-i);
		if (it==diags.end())
			return holder = T(0);
		if (j>i)
			return it->second[i-1];
		else
			return it->second[j-1];
	}
	T&operator()(int i, int j) 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=N);
		assert(j<=N);
		typename map<int, T*>::iterator it = diags.find(j-i);
		if (it==diags.end())
			return holder=T(0);
		if (j>i)
			return it->second[i-1];
		else
			return it->second[j-1];
	}
	void ZeroMatrix()
	{
		typename map<int, T*>::iterator it;
		for (it = diags.begin(); it!=diags.end(); it++)
		{
			T* h = it->second;
			for (int i = 0; i < (N-abs(it->first)); i++)
			{
				h[i] = T(0);
			}
		}
	}
	
	virtual void MulMatVec(Vector<T, Real> &R, const Vector<T, Real> &X) const
	{
		R.ZeroVector();
		typename map<int, T*>::const_iterator it;
		for(it = diags.begin(); it!=diags.end(); it++)
		{
			int d = it->first;
			int cnt = N - abs(d);
			const T* sx;
			T* sr;
			if (d>=0)
			{
				sx = X.GetPtr() + d;
				sr = R.GetPtr();
			}
			else
			{
				sx = X.GetPtr();
				sr = R.GetPtr() - d;
			}
			typename SWVector<T, Real>::constant XX(sx, cnt);
			typename SWVector<T, Real>::constant AA(it->second, cnt);
			SWVector<T, Real> RR(sr, cnt);
			VecMulAdd(RR, XX, AA);			
		}
	}


};
template<class T, class Real = T>
class SymDiagMatrix :public DiagMatrix<T, Real>
{
  protected:
  using DiagMatrix<T, Real>::diags;
  using DiagMatrix<T, Real>::N;
  using DiagMatrix<T, Real>::holder;
public:
	SymDiagMatrix(const vector<int> &v, int n):DiagMatrix<T, Real>(n)
	{
		for (int i = 0; i < v.size(); i++)
		{
			if(v[i]>=0)
			diags[v[i]] = new T[n-abs(v[i])];
		}
	}
	
	void AddDiag(int diag)
	{
		if (diag >=0)
			diags[diag] = new T[N-abs(diag)];
	}
	T operator()(int i, int j) const 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=N);
		assert(j<=N);
		if (i>j)
		{
			int p = j;
			j = i;
			i = p;
		}
		typename map<int, T*>::const_iterator it = diags.find(j-i);
		if (it==diags.end())
			return holder = T(0);
		if (j>i)
			return it->second[i-1];
		else
			return it->second[j-1];
	}
	T&operator()(int i, int j) 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=N);
		assert(j<=N);
		if (i>j)
		{
			return holder = T(0);
		}
		typename map<int, T*>::iterator it = diags.find(j-i);
		if (it==diags.end())
			return holder=T(0);
		if (j>i)
			return it->second[i-1];
		else
			return it->second[j-1];
	}
	virtual void MulMatVec(Vector<T, Real> &R, const Vector<T, Real> &X) const
	{
		R.ZeroVector();
		typename map<int, T*>::const_iterator it;
		for(it = diags.begin(); it!=diags.end(); it++)
		{
			int d = it->first;
			int cnt = N - abs(d);
			const T* sx;
			T* sr;
			sx = X.GetPtr() + d;
			sr = R.GetPtr();
			{
				typename SWVector<T, Real>::constant XX(sx, cnt);
				typename SWVector<T, Real>::constant AA(it->second, cnt);
				SWVector<T, Real> RR(sr, cnt);
				VecMulAdd(RR, XX, AA);			
			}
			if (d>0)
			{

				sx = X.GetPtr();
				sr = R.GetPtr() + d;

				typename SWVector<T, Real>::constant XX(sx, cnt);
				typename SWVector<T, Real>::constant AA(it->second, cnt);
				SWVector<T, Real> RR(sr, cnt);
				VecMulAdd(RR, XX, AA);	

			}
		}
	}

};

/* Uses virtual function to multiply matrix & vector */
template<class T, class Real>
void MulMatVec(Vector<T, Real> &R, const Matrix<T, Real> &A, const Vector<T, Real> &X)
{
	assert(A.Width()==X.Size());
	assert(A.Height()==R.Size());
	A.MulMatVec(R, X);
}

/*
Represent a matrix without own storage.
Could be better implemented (faster)
*/
template<class T, class Real = T>
class SubMatrix:public Matrix<T, Real>
{
	const int width;
	const int height;
	const int StartI;
	const int StartJ;
	Matrix<T, Real>& A;
	SubMatrix();
public:
	SubMatrix(Matrix<T, Real> &mat, int si, int sj, int rows, int columns)
		:A(mat), StartI(si-1), StartJ(sj-1), width(columns), height(rows)
	{
		assert(rows>0);
		assert(columns>0);
		assert(si>0);
		assert(sj>0);
		assert(rows+StartI <= mat.Height());
		assert(columns+StartJ <= mat.Width());
	}
	int Width() const 
	{
		return width;
	}
	int Height() const
	{
		return height;
	}
	T operator()(int i, int j) const 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=height);
		assert(j<=width);
		return A(i+StartI, j+StartJ);
	}
	T&operator()(int i, int j) 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=height);
		assert(j<=width);
		return A(i+StartI, j+StartJ);
	}
};

template<class T, class D>
void Transponate(Matrix<T, D> &res, Matrix<T, D>const&src)
{
	for (int i = 1; i <= src.Height(); i++)
		for (int j = 1; j<= src.Width(); j++)
			res(j, i) = src(i, j);
}

template<class T, class D>
void MultiplyMatrix(Matrix<T, D>& res, Matrix<T, D>const & a, Matrix<T, D>const &b)
{
	assert(!	 ( (res.Width() != b.Width() ) ||
			(res.Height()!=a.Height())||
			(a.Width()!=b.Height()) ));
		
	for (int i = 1; i <= a.Height(); i++)
		for (int j = 1; j <= b.Width(); j++)
		{
			T sum = 0;
			for (int k = 1; k <= b.Height(); k++)
				sum+=a(i, k)*b(k, j);
			res(i, j) = sum;
		}
}

template<class T, class D>
void Copy(Matrix<T, D>&a, const Matrix<T, D>&b)
{
	for (int i = 1; i <= a.Height(); i++)
		for (int j = 1; j <= a.Width(); j++)
			a(i, j) = b(i, j);
}

template<class T, class D>
void Copy(Vector<T, D>&a, const Vector<T, D> &b)
{
	assert(a.Size() == b.Size());
	T *ap = a.GetPtr();
	T const *bp = b.GetPtr();
	for (int i = 0; i < a.Size(); i++)
	{
		ap[i] = bp[i];
	}
}

template<class T, class D, class E>
void MulMatrixNum(Matrix<T, D>&res, const Matrix<T, D>&a, const E&b)
{
	for(int i = 1; i <= res.Height(); i++)
		for (int j = 1; j <= res.Width(); j++)
			res(i,j) = a(i,j) * b ;
}

template<class T, class D>
void MatrixSubstract(Matrix<T, D>&res, const Matrix<T, D>&a, const Matrix<T, D>&b)
{
	for(int i = 1; i <= res.Height(); i++)
		for (int j = 1; j <= res.Width(); j++)
			res(i,j) = a(i,j) - b(i,j);
}

/* 
Represent a matrix, Memory for all elements is reserved.
No Copy/assignement operators..
*/
template<class T, class Real = T>
class FullMatrix:public Matrix<T, Real>
{
	const int width;
	const int height;
	T *buf;

public:
	FullMatrix(int i, int j)
		:width(j), height(i), buf(new T[i*j])
	{
		assert(i>0);
		assert(j>0);
	}
	~FullMatrix()
	{
		delete[] buf;
	}
	int Width() const
	{
		return width;
	}
	int Height() const
	{
		return height;
	}
	T operator()(int i, int j) const 
	{
		assert(i>0);
		assert(j>0);
		assert(i<=height);
		assert(j<=width);
		return buf[i*width+j -width - 1];
	}
	T&operator()(int i, int j)
	{
		assert(i>0);
		assert(j>0);
		assert(i<=height);
		assert(j<=width);
		return buf[i*width+j -width - 1];
	}

};



template<class T, class D>
void ReverseMatrix(Matrix<T,D>&res, Matrix<T,D>&a)
{
	int d = a.Width();
	res(1,1) = T(1)/a(1,1);
	for (int r = 2; r <= d; r++)
	{
		SubMatrix<T,D>v(a, 1, r, r-1, 1);
		SubMatrix<T,D>u(a, r, 1, 1, r-1);
		SubMatrix<T,D>y(res, 1, r, r-1, 1);
		SubMatrix<T,D>x(res, r, 1, 1, r-1);

		FullMatrix< T , D > xt( 1 , r-1);
		SubMatrix<T,D>B(res, 1, 1, r-1, r-1);
		T &ann = a(r,r);
		T &alpha = res(r,r);
		FullMatrix<T,D> An1(r-1, r-1);
		FullMatrix<T,D> one(1,1);

		Copy(An1, B);
		MultiplyMatrix(xt, u, An1); // xt is now u*An1
		MultiplyMatrix(one, xt, v);
		MulMatrixNum(x, xt, alpha = T(1)/(one(1,1)-ann)); // we now have x;
		MultiplyMatrix(y, An1, v);
		MultiplyMatrix(B, y, x);
		MatrixSubstract(B, An1, B);
		MulMatrixNum(y, y, alpha);
		alpha*=-1;
	} 
}


/*print matrix using asterics...
*/
template<class T, class G>
void PrintMatrix1(const Matrix<T,G>&a)
{
	for (int i = 1; i <=a.Height();i++)
	{
		for (int j = 1; j <=a.Width();j++)
			cout << setw(2) << ((a(i,j)==0)?"  ":"*");
		cout << '\n';
	}
}

/* Prints matrix..
*/
template<class T, class G>
void PrintMatrix(const Matrix<T,G>&a)
{
	for (int i = 1; i <=a.Height();i++)
	{
		for (int j = 1; j <=a.Width();j++)
			cout <<setw(4)<< a(i, j);
		cout << "\n\n";
	}
}

/*
represents a 2d point
*/
template <class T>
class Point2d
{
public:
	Point2d(const Point2d&a): x(a.x), y(a.y){}
	Point2d(const T&X, const T&Y): x(X), y(Y){}
	Point2d(): x(0), y(0){}
	T x;
	T y;
};

/*
Numbers of the vercites of the triangle 
*/
class MeshTriangle
{
public:
	unsigned int v[3];
};

inline int GetNodeCount(int N)
{
	return (2*N-1)*N- N +1;
}

inline int GetNodeCountWithBounds(int N)
{
	return GetNodeCount(N)+4*N;
}




inline istream& operator>> (istream &in, Base::MeshTriangle &a)
{
	in >> a.v[0] >> a.v[1] >> a.v[2];
	return in;
}

inline istream& operator>> (istream &in, vector<Base::MeshTriangle> &a)
{
	a.clear();
	int n;
	in >> n;
	n *= 4*n;
	for (int i = 0; i < n; i++)
	{
		Base::MeshTriangle t;
		in >> t;
		a.push_back(t);
	}
	return in;
}
template<class T>
istream& operator >> (istream &in, Base::Point2d<T> &a)
{
	int n;
	in >> n >> a.x >> a.y;
	return in;
}

template<class T>
istream& operator >> (istream &in, vector<Base::Point2d<T> > &v)
{
	v.clear();
	int n;
	in >> n;
	int max = GetNodeCountWithBounds(n);
	for (int i = 0; i < max; i++)
	{
		Base::Point2d<T> a;
		in >> a;
		v.push_back(a);
	}
	return in;
}

inline ostream& operator << (ostream &out, const vector<Base::MeshTriangle > &a)
{
	out << (int) (sqrt((double)a.size())/2) << '\n';
	for(unsigned int i = 0; i < a.size(); i++)
		out << a[i].v[0] << ' ' << a[i].v[1] << ' ' << a[i].v[2] << '\n';
	return out;
}

template<class T>
ostream& operator << (ostream & out, const vector<Base::Point2d<T> > &v)
{
	int n = (int)(sqrt((v.size()-1)*8 + 4) - 2)/4;
	out << n << '\n';
	int prec = out.precision();
	out.precision(16);
	std::ios::fmtflags flags = out.flags();
	out.flags(flags | ios::left);
	for (int i = 0; i <v.size(); i++)
	{
		out <<setw(9)<< i + 1 << setw(22) << v[i].x << setw(22) << v[i].y << '\n';
	}
	out.precision(prec);
	out.flags(flags);
	return out;
}

template<class T, class Real>
ostream& operator << (ostream &out, const Base::DiagMatrix<T, Real> &a)
{
	const map<int, T*> &dia = a.GetDiags();
	out << a.Width() << '\n';
	out << dia.size() << '\n';
	typename map<int, T*>::const_iterator it;
	for (it = dia.begin(); it!=dia.end(); it++)
	{
		out << it->first << '\n';
		for (int i = 0; i< a.Width() - abs(it->first); i++)
			out	 << it->second[i] << '\n';
	}
	return out;
}

template<class T, class Real>
istream& operator >> (istream &in, Base::DiagMatrix<T, Real> &a)
{
	int n, nn;
	in >> n;
	a.NewMtx(n);
	in >> nn;
	for (int i = 0; i < nn; i++) // n is diag count
	{
		int dia;
		in >> dia;
		a.AddDiag(dia);
		T* p = (dia>=0)? &(a(1, dia+1)):&(a(1-dia, 1));
		for (int k = 0; k < n - abs(dia); k++)
			in >> *p++;
	}

	return in;
}


inline istream& operator >(istream &in, int& a)
{
	in.read((char*)&a, sizeof(a));
	return in;
}
inline istream& operator >(istream &in, double & a)
{
	in.read((char*)&a, sizeof(a));
	return in;
}
inline istream& operator >(istream &in, float & a)
{
	in.read((char*)&a, sizeof(a));
	return in;
}
inline ostream& operator <(ostream &out, const int& a)
{
	out.write((char*)&a, sizeof(a));
	return out;
}



inline istream& operator> (istream &in, Base::MeshTriangle &a)
{
	in.read((char*)&a.v[0], sizeof(a.v[0]));
	in.read((char*)&a.v[1], sizeof(a.v[1]));
	in.read((char*)&a.v[2], sizeof(a.v[2]));
	return in;
}
inline istream& operator> (istream &in, vector<Base::MeshTriangle> &a)
{
	a.clear();
	int n;
	in.read((char*)&n, sizeof(n));
	n *= 4*n;
	for (int i = 0; i < n; i++)
	{
		Base::MeshTriangle t;
		in > t;
		a.push_back(t);
	}
	return in;
}
template<class T>
istream& operator > (istream &in, Base::Point2d<T> &a)
{
	in.read((char*)&a.x, sizeof(a.x));
	in.read((char*)&a.y, sizeof(a.y));
	return in;
}

template<class T>
istream& operator > (istream &in, vector<Base::Point2d<T> > &v)
{
	v.clear();
	int n;
	in.read((char*)&n, sizeof(n));
	int max = GetNodeCountWithBounds(n);
	for (int i = 0; i < max; i++)
	{
		Base::Point2d<T> a;
		in > a;
		v.push_back(a);
	}
	return in;
}

inline ostream& operator < (ostream &out, const vector<Base::MeshTriangle > &a)
{
	int n = (int) (sqrt((double)a.size())/2); 
	out.write((char*)&n, sizeof(n));
	for(unsigned int i = 0; i < a.size(); i++)
	{
		out.write((char*)&a[i].v[0], sizeof(a[i].v[0]));
		out.write((char*)&a[i].v[1], sizeof(a[i].v[1]));
		out.write((char*)&a[i].v[2], sizeof(a[i].v[2]));
	}
	return out;
}

template<class T> 
ostream& operator< (ostream & out, const vector<Base::Point2d<T> > &v)
{
	int n = (int)(sqrt((v.size()-1)*8 + 4.) - 2)/4;
	out.write((char*)&n, sizeof(n));
	for (int i = 0; i <v.size(); i++)
	{
		out.write((char*)&v[i].x, sizeof(v[i].x));
		out.write((char*)&v[i].y, sizeof(v[i].y));
	}
	return out;
}

template<class T, class Real>
ostream& operator < (ostream &out, const Base::DiagMatrix<T, Real> &a)
{
	const map<int, T*> &dia = a.GetDiags();
	int aa = a.Width();
	out.write((char*)&aa, sizeof(aa));
	aa = dia.size();
	out.write((char*)&aa, sizeof(aa));
	typename map<int, T*>::const_iterator it;
	for (it = dia.begin(); it!=dia.end(); it++)
	{
		out.write((char*)&it->first, sizeof(it->first));
		for (int i = 0; i< a.Width() - abs(it->first); i++)
			out.write((char*)&it->second[i], sizeof(it->second[i]));
	}
	return out;
}

template<class T, class Real>
istream& operator>(istream &in, Base::FullVector<T, Real> &a)
{
	int size;
	in > size;
	a.SetSize(size);
	in.read((char*)a.GetPtr(), a.Size()*sizeof(T));
	return in;
}

template<class T, class Real>
ostream& operator<(ostream &out, Base::FullVector<T, Real> &a)
{
	out < a.Size();
	out.write((char*)a.GetPtr(), a.Size()*sizeof(T));
	return out;
}

template<class T, class Real>
istream& operator>(istream &in, Base::DiagMatrix<T, Real> &a)
{
	int n, nn;
	in > n;
	a.NewMtx(n);
	in > nn;
	for (int i = 0; i < nn; i++) // n is diag count
	{
		int dia;
		in > dia;
		a.AddDiag(dia);
		T* p = (dia>=0)? &(a(1, dia+1)):&(a(1-dia, 1));
		for (int k = 0; k < n - abs(dia); k++)
			in > *p++;
	}

	return in;
}

} // namespace Base;

} // My
#endif

