#include <iostream>
#include <stdexcept>
#include <fstream>
#include <map>
#include <iomanip>
#include <vector>


#include "string_util.h"

#include "smallops.h"

template<class T>
void PutInRange(T &a, const T&l, const T&r)
{
  if (a <l)
    a = l;
  if (a>r)
    a = r;
}


template<class Double>
class TwoValue
{
  double CurTimeA[2];
  double alpha[2];
  int o1, o2;

  int NX, NY;
  Double SizeX, SizeY;
  Double StartX, StartY;
  Double EndX, EndY;
  std::ifstream f;
  std::vector<Double> data;
public:
  TwoValue()
  {
    alpha[0] = alpha[1] = 0.0;
    CurTimeA[0] = CurTimeA[1] = -1;
    o1 = 0;
    o2 = 1;
  } 
  void SetFilename(const std::string &fname)
  {
    f.open(fname.c_str());
    if (!f)
      throw std::runtime_error("Cannot open file: " + fname);
    ScanParameters();
  }
  // gets the dimentions
  // reads the minimum and maximum values, sizes and coordnates

  void ParseCoordsLine(int &N, Double &start, Double &end, Double &size)
  {
    double prev_coord = 0.0;
    double cur_coord = 0.0;
    std::string line;
    getline(f, line);
    std::istringstream f(line);
    N = 0;
    while(f)
    {
      double a;
      f >> a;
      if (f)
      {
	prev_coord = cur_coord;
	cur_coord = a;
	N++;
      }	
    }

    end = cur_coord;
    size = cur_coord-prev_coord;
    start = cur_coord - size*(N-1);
  }
    
  void ScanParameters()
  {
    std::string line;
    getline(f, line);
    std::string l,r;
    Split(line, l, r, "% ");
    Trim(r);
    if(r != "Grid")
      throw std::runtime_error("Grid not found in file");
    ParseCoordsLine(NX, StartX, EndX, SizeX);
    ParseCoordsLine(NY, StartY, EndY, SizeY);      
  }
  void RollTime(Double Time)
  {
    if (o1==o2) // we have already reached the end of time
      return;
    while(Time >=CurTimeA[o2]) // time is less the newer time
    {
      if(!ReadNextTime())
      {
	o1 = o2;
	alpha[0] = 1.0;
	alpha[1] = 0.0;
	return;
      }
      // else
      if (CurTimeA[o1] <0) 
	continue;// we read 2 times in the beggining
      
      //      std::cout << CurTimeA[o1] <<" <--> " << CurTimeA[o2] << std::endl;
    }
    Double time_diff = CurTimeA[o2] - CurTimeA[o1];
    alpha[o1] = 1.-(Time - CurTimeA[o1])/time_diff;
    alpha[o2] = 1.-(CurTimeA[o2] - Time)/time_diff;
  }

  Double ReadSingleValue()
  {
    std::string str;
    f >> std::skipws >>str;
    Double res = Convert<Double>(str);
    //    std::cout << str << " _______ " <<res << std::endl;
    return res;
  }
    
  // reads data; o1 holds older data, o2 holds newest data
  void ReadData(double NewTime)
  {
    // change the data size
    data.resize(2ul*NX*NY);
    // we first read into o1
    for (size_t i = 0; i <data.size(); i+=2)
    {
      data[i+o1] = ReadSingleValue();
    }
    CurTimeA[o1] = NewTime;   
    o1 = 1-o1;
    o2 = 1-o2;
  }
  
  bool ReadNextTime()
  {
    Double NextTime = -1;
    std::string line;
    while (true){
      getline(f, line);
      if (!f)
	return false;
      std::string l, r;
      Split(line, l, r, "Time=");
      if (r != "")
      {
	NextTime = Convert<Double>(r);
	break;
      }
    }
    ReadData(NextTime);
    return true;
  }


  // gets an interpolated value within range
  Double GetValue(int x, int y, int off)
  {
    return data[off + (y*NX+x)*2];
  }
  Double GetValue(Double X, Double Y)
  {
    int Ix = (X-StartX)/SizeX;
    int Iy = (Y-StartY)/SizeY;
    PutInRange(Ix, 0, NX-2);
    PutInRange(Iy, 0, NY-2);
    Double cx[2];
    cx[1] = X-Ix*SizeX;
    cx[0] = 1.-cx[1];
    
    Double cy[2];
    cy[1] = Y-Iy*SizeY;
    cy[0] = 1.-cy[1];
    Double tempres[2];
    tempres[0] = tempres[1] = 0.0;
    for (int xi = 0; xi<2; ++xi)
      for (int yi = 0; yi<2; ++yi)
	for (int off = 0; off<2; ++off)
	  tempres[off] += GetValue(Ix+xi, Iy+yi, off)*cx[xi]*cy[yi];
    
    Double res = tempres[0] * alpha[0] + tempres[1]*alpha[1];
    //    std::cout << alpha[o1] << " -------- " << alpha[o2] << std::endl;
    return res;
  
  }
  
  Double operator()(Double X, Double Y)
  {
    PutInRange(X, StartX, EndX);
    PutInRange(Y, StartY, EndY);
    return GetValue(X, Y);
  }
    
};

template<class Double>
class ComsolData{
  Double StepX, StepY;
  Double MinX, MaxX;
  Double MinY, MaxY;
  std::map<std::string, TwoValue<Double> > data;
};


int main(int argc, char ** argv)
{
  std::string fname = "file";
  if (argc>1)
  {
    fname = argv[1];
  }
  TwoValue<double> val;
  try{
    val.SetFilename(fname);
    for (double time = 0.0; time < 800.; time += 4.5)
    {
      val.RollTime(time);
      std::cout << std::setw(10) << time << " " << val(0.0106, 0.0169) << std::endl;
    }
  }catch(std::exception &e)
  {
    std::cout <<"Caught error: " << e.what() << std::endl;
  }
}
