#include "common.h"

#include "mesh.h"
#include "ellipt.h"
#include "pcg_mic0.h"
#include "timer.h"
#include "visual.h"

template <class T, bool Sym>
void print_matrix(const sparse_matrix<T, Sym> & A)
{
    size_t N = A.size();
    for (size_t k = 0; k < N; ++k) {
        for (typename sparse_matrix<T, Sym>::row_iterator
                 it = A.row_begin(k),
                 end = A.row_end(k); it != end; ++it) {
            std::cout << it.column() << " -> ";
            std::cout << it.value() << "\t\t";
        }
        std::cout << std::endl;
    }
}

int main(int argc, char * argv[])
try {
    glutInit(&argc, argv);
    if (argc < 2) 
        throw std::runtime_error("usage: visual <meshname> [<refinement>]");
    size_t ref_n = 0;
    if (argc > 2) 
        std::istringstream (argv[2]) >> ref_n;
    mesh_2d<double> mesh;
    std::cout << "loading mesh" << std::endl;
    {
        simple_timer timer;
        mesh.load(argv[1]);
    }
    std::cout << "refining mesh" << std::endl;
    {
        simple_timer timer;
        mesh.refine(ref_n);
    }
    std::cout << "renumbering" << std::endl;
    {
        simple_timer timer;
        mesh.renumber();
    }
    size_t N = mesh.nodes.size();
    asm_matrix<double, true> Aa(N);
    vector<double> f(N);
    vector<double> u(N);
    std::cout << "assembling stiffness matrix" << std::endl;
    {
        simple_timer timer;
        asm_global_matrix_t3(Aa, f, mesh);
    }
    sparse_matrix<double, true> A(Aa);
    Aa.clear();

    std::cout << "solving linear system" << std::endl;
    size_t iter;
    try {
        simple_timer timer;
        iter = pcg(u, A, f, 1e-10);
        std::cout << iter << " iterations" << std::endl;
    } catch (const mic0_failure & fail) {
        std::cout << "MIC(0) failed" << std::endl;
    }

     foreach(const mesh_node_2d<double> & n, mesh.nodes) {
         u(n.number) -= U(n.x, n.y);
         u(n.number) = std::fabs(u(n.number));
     }
    
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutCreateWindow("Plot");
    glutDisplayFunc(display);
    glutCreateMenu(menu);
    glutAddMenuEntry("Save EPS", SAVE_EPS);
    glutAddMenuEntry("Exit", QUIT);
    glutAttachMenu(GLUT_RIGHT_BUTTON);
    gfxinit(mesh, u);
    glutMainLoop();
} catch (const std::runtime_error & e) {
    std::cerr << e.what() << std::endl;
}
