#ifndef __ELLIPT_H__
# define __ELLIPT_H__

# include "common.h"
# include "matrix_vector.h"
# include "mesh.h"
# include "pcg_mic0.h"

template <class T> T U(T x, T y)
{
    return x*x*x*x + y*y*y;
}

const double Q = 2.3;

template <class T> T F(T x, T y)
{
  return -12*x*x - 6*y + Q*U(x,y);
}

// dirichlet conditions
template <class T> T dirichlet(T x, T y)
{
    return U(x, y);
}

template<class T, int N>
void MulMat(T R[N][N], const T A[N][N], const T B[N][N])
{
  for(int i = 0; i < N; ++i)
    for(int j = 0; j < N; ++j)
      {
        T sum = 0;
        for(int k = 0; k < N; k++)
          sum += A[i][k] * B[k][j];
        R[i][j] = sum;
      }
}
template<class T, int N>
void Transpose(T A[N][N], const T B[N][N])
{
  for(int i = 0; i < N; ++i)
    for(int j = 0; j < N; ++j)
      A[j][i] = B[i][j];
}
template<class T, int N>
inline void GausJordanInverse(T R[N][N],  T A[N][N])
{
  std::vector<int> ind(N);

  typedef T Double;

  Double max, max1;
  for(int i = 0; i < N; ++i)
    {
      ind[i] = i;
      for(int k = 0; k<N; ++k)R[i][k] = 0;
      R[i][i] = 1;
    }
  int imax;
  
  for(int i = 0; i < N; ++i)
    {
      imax = i;
      max = fabs(A[i][ind[i]]);
      //find pivot element on i-th row;
      for(int k = i; k < N; ++k)
	if(fabs(A[i][ind[k]])>max)
	  {
	    imax = k;
	    max = fabs(A[i][ind[k]]);
	  }
      std::swap(ind[i], ind[imax]);
      
      max = 1/A[i][ind[i]];
      for(int k = 0; k < N; ++k)
	{
	  A[i][k] *= max;
	  R[i][k] *= max;
	}

      for (int j = 0; j < N; ++j)
	{
	  if(j != i)
	    {
	      max = A[j][ind[i]];
 	      for(int k = 0; k < N; ++k)
 		{
 		  A[j][ind[k]] -= max*A[i][ind[k]];
 		  R[j][ind[k]] -= max*R[i][ind[k]];
 		}
	    }
	}
    }
  for(int i = 0; i < N; i++)
    {
      if(ind[i]!=i)
	{
	  for(int k = 0; k < N; ++k)
	    {
	      std::swap(A[i][k], A[ind[i]][k]);
	      std::swap(R[i][k], R[ind[i]][k]);
	    }
	  std::swap(ind[i], ind[ind[i]]);
	}
    }
      
}



template <class T> void element_matrix_t3(T Ae[3][3], T fe[6],
                                          T x1, T y1, T x2, T y2, T x3, T y3,
                                          T A, T B, T C)
{
  using namespace std;
    // xij = xi - xj
    T x31 = x3 - x1; T x21 = x2 - x1; T x32 = x3 - x2; 
    T y31 = y3 - y1; T y21 = y2 - y1; T y32 = y3 - y2;

    T M[3][3];
    T f[3];


    M[0][0] = 1./12.;
    M[0][2] = M[0][1] = M[2][0] = M[1][0] = 1./24.;
    M[1][1] = M[2][2] = 1./12.;
    M[2][1] = M[1][2] = 1./24.;
    
		//    MulMat(tmp, M, H);
		//    MulMat(M, Ht, tmp);





    // jacobian
    T J = x21*y31 - x31*y21;

    // element stiffness matrix
    T D = 2*std::abs(J);


    
    Ae[0][0] = (C*x32*x32 + y32*(A*y32 - 2*B*x32))/D;
    Ae[0][1] = (-C*x31*x32 + B*(y31*x32 + x31*y32) - A*y31*y32)/D;
    Ae[0][2] = (C*x21*x32 - B*(y21*x32 + x21*y32) + A*y21*y32)/D;
    Ae[1][0] = Ae[0][1];
    Ae[1][1] = (C*x31*x31 + y31*(A*y31 - 2*B*x31))/D;
    Ae[1][2] = (-C*x21*x31 + B*(y21*x31 + x21*y31) - A*y21*y31)/D;
    Ae[2][0] = Ae[0][2];
    Ae[2][1] = Ae[1][2];
    Ae[2][2] = (C*x21*x21 + y21*(A*y21 - 2*B*x21))/D;



    for(int i = 0; i < 3; ++i)
      {
        for(int j = 0; j < 3; ++j)
          {
	    M[i][j]*=J;
            Ae[i][j] += M[i][j] * Q;
          }
       }

    

     D = std::abs(J);
     fe[0] = (F(x1, y1)/6)*D;
     fe[1] = (F(x2, y2)/6)*D;
     fe[2] = (F(x3, y3)/6)*D;
}

template <class T> void apply_dirichlet(asm_matrix<T, true> & A,
                                        vector<T> & f,
                                        const mesh_2d<T> & mesh)
{
    size_t N = mesh.nodes.size();
    foreach(const mesh_node_2d<T> & n, mesh.nodes) {
        if (0 < n.boundary) {
            T d = dirichlet(n.x, n.y);
            for (size_t i = N; i--;) {
                f(i) -= d*A(i, n.number);
                A(i, n.number) = T();
                A(n.number, i) = T();
            }
            A(n.number, n.number) = 1.;
            f(n.number) = d;
        }
    }
}

template <class T> void apply_neumann(asm_matrix<T, true> & A,
                                      vector<T> & f,
                                      const mesh_2d<T> & mesh)
{
    foreach (const typename mesh_2d<T>::poly_set::value_type & e1,
             mesh.poly) {
        const mesh_node_2d<T> & n1 = mesh.node(e1.first);
        foreach (const typename mesh_2d<T>::poly_row::value_type & e2,
                 e1.second) {
            const mesh_node_2d<T> & n2 = mesh.node(e2.first);
            size_t bound = e2.second;
            T h = sqrt((n1.x - n2.x)*(n1.x - n2.x) +
                       (n1.y - n2.y)*(n1.y - n2.y));
//             f(n1.number) += h*px[bound];
//             f(n2.number) += h*px[bound];
        }
    }
}

template <class T> void asm_global_matrix_t3(asm_matrix<T, true> & A,
                                             vector<T> & f,
                                             const mesh_2d<T> & mesh)
{
    foreach (const mesh_element_2d<T> & e, mesh.elements) {
        T x[3], y[3];
        size_t num[3];
        for (int i = 3; i--;) {
            mesh_node_2d<T> n = mesh.node(e.nodes[i]);
            x[i] = n.x; y[i] = n.y; num[i] = n.number;
        }
        T Ae[3][3], b[3];
        size_t reg = e.region();
        element_matrix_t3(Ae, b, x[0], y[0], x[1],
                          y[1], x[2], y[2], 1., 0., 1.);
        for (int i = 3; i--;) {
            for (int j = 3; j--;)
                A(num[i], num[j]) += Ae[i][j];
            f(num[i]) += b[i];
        }
    }
    apply_neumann(A, f, mesh);
    apply_dirichlet(A, f, mesh);
}

#endif 
