#ifndef __functiontricubic_h__
#define __functiontricubic_h__
#include "small.h"
#include "mirorer.h"
#include <cassert>
#include <cmath>


template<class ValueT, class IndexT = ValueT, class StorageT=ValueT>
class TricubicFunction
{

  static double P(double t, double const (&a)[4])
  {
    double mat[4][4] = {
      { 0,  2,  0, 0},
      {-1,  0,  1, 0},
      { 2, -5,  4, -1},
      {-1,  3, -3, 1}};
    double vec[4] = {0,0,0,0};
    for (int i = 0; i < 4; ++i)
      for (int j = 0; j < 4; ++j)
	vec[i] += mat[i][j]*a[j];
    double T[4] = {1, t, t*t, t*t*t};
    double res = 0;
    for (int i = 4; i--;)
      res += T[i]*vec[i];
    return res*0.5;    
  }
public:
  typedef ValueT value_type;
  typedef IndexT index_type;
  typedef SmallVector<index_type,3> P3D;
  vector3d<StorageT> buf;
  P3D Imin, Imax;
  P3D unit_size;
  P3D range;
  P3D zero;
  P3D zero1;
  template<class T>
  void Init(const vector3d<T>&v,
	    P3D const &min=0, P3D const &max=1, bool extrapolate = true)
  {
    Imin = min;
    Imax = max;
    range = Imax;
    range -= Imin;
    unit_size = range;

    for(int i = 0; i < 3; ++i)
      unit_size[i]/=v.sizes[i];
    buf.SetSize(v.sizes[0]+3, v.sizes[1]+3, v.sizes[2]+3);


    zero1 = unit_size;
    
    zero1 /= 2.;
    zero =  min;
    zero -= zero1;
    typedef typename vector3d<T>::size_type Int;
    for(Int i = 0; i<buf.sizes[0]; ++i)
      for(Int j = 0; j<buf.sizes[1]; ++j)
	for(Int k = 0; k <buf.sizes[2];++k)
	  buf(i, j, k) = 0;

    for(Int i = 0; i <v.sizes[0]; ++i)
    {
      for(Int j = 0; j < v.sizes[1]; ++j)
      {
	for(Int k = 0; k < v.sizes[2];++k)
	{
	    buf(i+1, j+1, k+1) = v(i,j,k);
	}
      }
    }
    if(extrapolate)
    {
      for(Int i = 0; i < buf.sizes[0]; ++i)
	for(Int j = 0; j < buf.sizes[1]; ++j)
	{
	  const Int N = buf.sizes[2] - 2;
	  buf(i,j,0) = 2*buf(i,j,1) - buf(i,j,2);
	  buf(i,j,N) = 2*buf(i,j,N-1) - buf(i,j,N-2);
	  buf(i,j,N+1) = buf(i,j,N);
	}
      for(Int i = 0; i < buf.sizes[0]; ++i)
	for(Int k = 0; k < buf.sizes[2]; ++k)
	{
	  const Int N = buf.sizes[1] - 2;
	  buf(i,0,k) = 2*buf(i,1,k) - buf(i,2,k);
	  buf(i,N,k) = 2*buf(i,N-1,k) - buf(i,N-2,k);
	  buf(i,N+1,k) = buf(i, N, k);
	}
      for(Int j = 0; j < buf.sizes[1]; ++j)
	for(Int k = 0; k < buf.sizes[2]; ++k)
	{
	  const Int N = buf.sizes[0] - 2;
	  buf(0,j,k) = 2*buf(1,j,k) - buf(2,j,k);
	  buf(N,j,k) = 2*buf(N-1,j,k) - buf(N-2, j,k);
	  buf(N+1,j,k) = buf(N, j, k);
	}
    }
    else
    {
      for(Int i = 0; i < buf.sizes[0]; ++i)
	for(Int j = 0; j < buf.sizes[1]; ++j)
	{
	  const Int N = buf.sizes[2] - 2;
	  buf(i,j,0) = 0;//1*buf(i,j,1);// - buf(i,j,2);
	  buf(i,j,N) = 0;//1*buf(i,j,N-1);// - buf(i,j,N-2);
	  buf(i,j,N+1) = 0;// buf(i,j,N);
	}
      for(Int i = 0; i < buf.sizes[0]; ++i)
	for(Int k = 0; k < buf.sizes[2]; ++k)
	{
	  const Int N = buf.sizes[1] - 2;
	  buf(i,0,k) = 0;//1*buf(i,1,k);// - buf(i,2,k);
	  buf(i,N,k) = 0;//1*buf(i,N-1,k);// - buf(i,N-2,k);
	  buf(i,N+1,k) = 0;//buf(i, N, k);
	}
      for(Int j = 0; j < buf.sizes[1]; ++j)
	for(Int k = 0; k < buf.sizes[2]; ++k)
	{
	  const Int N = buf.sizes[0] - 2;
	  buf(0,j,k) = 0;//1*buf(1,j,k);// - buf(2,j,k);
	  buf(N,j,k) = 0;//1*buf(N-1,j,k);// - buf(N-2, j,k);
	  buf(N+1,j,k) =0;// buf(N, j, k);
	}
    }
  }


  void CalcInd(P3D p, SmallVector<size_t, 3> &ind, P3D&c)const
  {
    //    std::cout <<p[0] << " " <<p[1] << " " << p[2]<< std::endl;
    //    std::cout <<x[0] << " " <<x[1] << " " << x[2]<< std::endl;
    //    std::cout <<ind[0] << " " <<ind[1] << " " << ind[2]<< std::endl;
    for(int i = 0; i < 3; ++i)
    {
      //      std::cout <<std::setw(10)<< p[i] << std::endl;
      if (p[i] <  (Imin[i] - zero1[i]))
        p[i] = Imin[i];
      assert(p[i] >  (Imin[i] - zero1[i]));
      if (p[i] >  (Imax[i] + zero1[i]))
      {
	p[i] = Imax[i];
	assert(p[i] <  (Imax[i] + zero1[i]));
      }
    }
    P3D x;
    x = p;
    x -= zero;
    for (int i = 0; i < 3; ++i)
    {
      ind[i] =(int) trunc(x[i]/unit_size[i]);
      if (ind[i] <1)
      {
	ind[i] = 1;
	c[i] = 0;
      }
      else
      if (ind[i]>(buf.sizes[i]-3))
      {
	ind[i] = buf.sizes[i]-3;
	c[i] = 1;
      }
      else
	c[i] = (x[i] - ind[i]*unit_size[i])/unit_size[i];
    }    
    
  }
  value_type RetValA(const SmallVector<size_t, 3>&ind,  const P3D &t, int j, int k)const
    
  {
    double v[4] = {
      buf(ind[0] - 1 ,ind[1] + j, ind[2] + k),
      buf(ind[0] + 0 ,ind[1] + j, ind[2] + k),
      buf(ind[0] + 1 ,ind[1] + j, ind[2] + k),
      buf(ind[0] + 2 ,ind[1] + j, ind[2] + k)
    };
    return P(t[0], v);      
  }
  value_type RetValB(const SmallVector<size_t, 3>&ind,  const P3D &t, int  k)const
  {
    double v[4] = {
      RetValA(ind, t, -1, k),
      RetValA(ind, t, 0, k),
      RetValA(ind, t, 1, k),
      RetValA(ind, t, 2, k)
    };
    return P(t[1], v);
  }
    
  
  value_type RetVal(const SmallVector<size_t, 3>&ind,  const P3D &t)const
  {
    double v[4] = {
      RetValB(ind, t, -1),
      RetValB(ind, t, 0),
      RetValB(ind, t, 1),
      RetValB(ind, t, 2)
    };

    //return RetValB(ind, t, 0);
    return P(t[2], v);
  }

  value_type operator()(const P3D &p)const
  {
    SmallVector<size_t, 3> ind;
    P3D c;
    CalcInd(p, ind, c);
    return RetVal(ind, c);
  }
  value_type operator()(index_type x, index_type y, index_type z)const
  {
    SmallVector<size_t, 3> ind;
    P3D c;
    P3D p;
    p[0] = x; p[1] = y; p[2] = z;
    CalcInd(p, ind, c);
    return RetVal(ind, c);
  }
};



#endif

