#ifndef MATRIX_VECTOR_H_
#define MATRIX_VECTOR_H_

#include "common.h"

//
// SPARSE MATRIX CLASSES
//

template <bool Sym> struct index_policy { };

template <> struct index_policy<true> {

    static bool can_set(size_t i, size_t j)
    {
        return i <= j;
    }
    
    static void sort(size_t & i, size_t & j)
    {
        if (i > j)
            std::swap(i, j);
    }
};

template <> struct index_policy<false> {

    static bool can_set(size_t i, size_t j)
    {
        return true;
    }
    
    static void sort(size_t & i, size_t & j) { }
};

// Assembly matrix
template <class T, bool Sym = false>
class asm_matrix {

 public:
    typedef std::map<size_t, T> row;

 private:
    std::vector<row> data;
	
	class element {

		asm_matrix & m;
		size_t i, j;

     public:
		element(asm_matrix & m, size_t i, size_t j)
          : m(m), i(i), j(j) 
		{
		}

        template <class V> element & operator=(V t)
		{
			m.set(i, j, t);
            return *this;
		}

#define OP(op) \
        template <class V> element & operator op##=(V t) \
        { \
            m.set(i, j, m.get(i, j) op t); \
            return *this; \
        }
        OP(+) OP(-) OP(*) OP(/)
#undef OP
		
		operator T() const
		{
			return m.get(i, j);
		}
	};

    void set(size_t i, size_t j, T t)
    {
        if (!index_policy<Sym>::can_set(i, j))
            return;
        if (T() == t) {
            if (T() == get(i, j)) {
                return;
            } else {
                data[i].erase(j);
                return;
            }
        }
        
        data[i][j] = t;
    }

    T get(size_t i, size_t j) const
    {
        index_policy<Sym>::sort(i, j);
        typename row::const_iterator it = data[i].find(j);
        return (it == data[i].end()) ? T() : it->second;
    }

 public:
	explicit asm_matrix(size_t N) : data(N)
	{
        for (int i = 0; i < N; ++i)
            data[i][i] = T();
	}

	element operator()(size_t i, size_t j)
	{
        return element(*this, i, j);
	}

    const element operator()(size_t i, size_t j) const
    {
        return element(*this, i, j);
    }

    row & irow(size_t i)
    {
        return data[i];
    }
    
    const row & irow(size_t i) const
    {
        return data[i];
    }
    
	size_t size() const
    {
        return data.size();
    }

    // don't try to use this class after you call clear() 
    void clear()
    {
        std::vector<row> empty;
        std::swap(data, empty);
    }
};

// compressed sparse row storage matrix class
template <class T, bool Sym = false> class sparse_matrix {

    typedef asm_matrix<T, Sym> assembly;
    
    std::vector<T> aa;
    std::vector<size_t> ja;
    std::vector<size_t> ia;
    
 public:

    explicit sparse_matrix(const assembly & m)
      : ia(m.size() + 1)
    {
        for (size_t i = 1; i < ia.size(); ++i)
            ia[i] = ia[i - 1] + m.irow(i - 1).size();
        size_t N = ia.back();
        ja.resize(N);
        aa.resize(N);
        size_t n = 0;
        for (size_t i = 0; i < m.size(); ++i) {
            const typename assembly::row & r = m.irow(i);
            for (typename assembly::row::const_iterator it = r.begin(),
                     end = r.end(); it != end; ++it) {
                ja[n] = it->first;
                aa[n] = it->second;
                ++n;
            }
        }
    }

    size_t size() const
    {
        return ia.size() - 1;
    }

    class row_iterator {

        const sparse_matrix & mat;
        size_t i;

     public:
        row_iterator(const sparse_matrix & mat, int i)
          : mat(mat), i(i)
        {
        }
        
        row_iterator & operator++()
        {
            ++i;
            return *this;
        }

        row_iterator operator++(int)
        {
            row_iterator tmp = *this;
            ++(*this);
            return tmp;
        }

        size_t column() const
        {
            return mat.ja[i];
        }

        const T & value() const
        {
            return mat.aa[i];
        }

        bool operator==(const row_iterator & o) const
        {
            return /*(&mat == &o.mat) && */(i == o.i);
        }

        bool operator!=(const row_iterator & o) const
        {
            return !(*this == o);
        }

        ptrdiff_t operator-(const row_iterator & o) const 
        {
            return i - o.i;
        }
    };

    row_iterator row_begin(size_t row) const
    {
        return row_iterator(*this, ia[row]);
    }

    row_iterator row_end(size_t row) const
    {
        return row_iterator(*this, ia[row + 1]);
    }
};

//
// VECTORS
//

template <class T> class vector;

// Base class for Vector expressions
template <class Derived, class T> struct v_expr {

    Derived & derived()
    {
        return reinterpret_cast<Derived &>(*this);
    }
    
    const Derived & derived() const
    {
        return reinterpret_cast<const Derived &>(*this);
    }

    // returns the size of the resulting vector
    size_t size() const
    {
        return derived().size();
    }

    // assigns the result to a vector
    void assign_to(vector<T> & r) const
    {
        derived().assign_to_impl(r);
    }

    // indexing operation
    T operator()(size_t i) const
    {
        return derived().operator()(i);
    }
};

// Default implementation of assing_to
template <class Derived, class T>
struct v_expr_with_assign : public v_expr<v_expr_with_assign<Derived, T>, T> {

    Derived & derived()
    {
        return reinterpret_cast<Derived &>(*this);
    }
    
    const Derived & derived() const
    {
        return reinterpret_cast<const Derived &>(*this);
    }
    
    size_t size() const
    {
        return derived().size();
    }

    void assign_to_impl(vector<T> & r) const
    {
        for (size_t i = size(); i--;)
            r(i) = (*this)(i);
    }

    T operator()(size_t i) const
    {
        return derived().operator()(i);
    }
};

// simple vector class
template <class T> class vector : public v_expr_with_assign<vector<T>, T> {

    std::vector<T> data;

 public:
    explicit vector(size_t N, T val = T()) : data(N, val)
    {
    }

    template <class V>
    vector(const v_expr<V, T> & e)
      : data(e.size())
    {
        e.assign_to(*this);
    }

    T & operator()(size_t i)
    {
        return data[i];
    }

    T operator()(size_t i) const
    {
        return data[i];
    }

    size_t size() const
    {
        return data.size();
    }

    template <class V>
    vector & operator=(const v_expr<V, T> & e)
    {
        e.assign_to(*this);
    }
};

//
// VECTOR EXPRESSIONS
//

// Sparse matrix by vector multiplication
template <class T, class V>
struct mv_mul : public v_expr_with_assign<mv_mul<T, V>, T> {

    typedef sparse_matrix<T, false> Matrix;
    
    const Matrix & m;
    const v_expr<V, T> & v;

    mv_mul(const Matrix & m, const v_expr<V, T> & v)
      : m(m), v(v)
    {
    }

    size_t size() const
    {
        return v.size();
    }

    T operator()(size_t i) const
    {
        T dp = 0;
        for (typename Matrix::row_iterator it = m.row_begin(i),
                 end = m.row_end(i); it != end; ++it) 
            dp += it.value()*v(it.column());
        return dp;
    }
};

template <class T, class V> 
inline mv_mul<T, V> operator*(const sparse_matrix<T, false> & m,
                              const v_expr<V, T> & v)
{
    return mv_mul<T, V>(m, v);
}

// Transpose matrix by vector multiplication
template <class T, class V>
struct mtv_mul : public v_expr<mtv_mul<T, V>, T> {

    typedef sparse_matrix<T, false> Matrix;
    
    const Matrix & m;
    const v_expr<V, T> & v;

    mtv_mul(const Matrix & m, const v_expr<V, T> & v)
      : m(m), v(v)
    {
    }

    size_t size() const
    {
        return v.size();
    }

    T operator()(size_t i) const
    {
        // it's not possible to use transpose matrix by vector
        // multiplication in an expression, so this must not compile
        return T::INVALID_OPERATION();
    }

    void assign_to_impl(vector<T> & r) const
    {
        for (size_t i = size(); i--;)
            r(i) = T();
        for (size_t i = size(); i--;) {        
            typename Matrix::row_iterator it = m.row_begin(i),
                end = m.row_end(i);
            for (; it != end; ++it) {
                size_t j = it.column();
                r(j) += it.value()*v(i);
            }
        }
    }
    
};

template <class T, class V> 
inline mtv_mul<T, V> transpose_mul(const sparse_matrix<T, false> & m,
                                   const v_expr<V, T> & v)
{
    return mtv_mul<T, V>(m, v);
}

// Symmetric matrix by vector multiplication
template <class T, class V>
struct sym_mv_mul : public v_expr<sym_mv_mul<T, V>, T> {

    typedef sparse_matrix<T, true> Matrix;
    
    const Matrix & m;
    const v_expr<V, T> & v;

    sym_mv_mul(const Matrix & m, const v_expr<V, T> & v)
      : m(m), v(v)
    {
    }

    size_t size() const
    {
        return v.size();
    }

    T operator()(size_t i) const
    {
        // it's not possible to use symmetric matrix by vector
        // multiplication in an expression, so this must not compile
        return T::INVALID_OPERATION();
    }

    void assign_to_impl(vector<T> & r) const
    {
        for (size_t i = size(); i--;) {        
            typename Matrix::row_iterator it = m.row_begin(i),
                end = m.row_end(i);
            T r_i = it.value()*v(i);
            for (; ++it != end;) {
                size_t j = it.column();
                r_i += it.value()*v(j);
                r(j) += it.value()*v(i);
            }
            r(i) = r_i;
        }
    }
    
};
    
template <class T, class V> 
inline sym_mv_mul<T, V> operator*(const sparse_matrix<T, true> & m,
                                  const v_expr<V, T> & v)
{
    return sym_mv_mul<T, V>(m, v);
}

// Scalar by vector multiplication
template <class T, class V>
struct sv_mul : public v_expr_with_assign<sv_mul<T, V>, T> {

    T s;
    const v_expr<V, T> & v;

    sv_mul(T s, const v_expr<V, T> & v)
      : s(s), v(v)
    {
    }

    size_t size() const
    {
        return v.size();
    }

    T operator()(size_t i) const
    {
        return s*v(i);
    }
};

template <class T, class V> 
inline sv_mul<T, V> operator*(T s, const v_expr<V, T> & v)
{
    return sv_mul<T, V>(s, v);
}

template <class T, class V> 
inline sv_mul<T, V> operator*(const v_expr<V, T> & v, T s)
{
    return sv_mul<T, V>(s, v);
}

// Vector inner product
template <class T, class V1, class V2>
inline T operator,(const v_expr<V1, T> & v1,
                   const v_expr<V2, T> & v2)
{
    T res = T();
    for (size_t i = v1.size(); i--;)
        res += v1(i)*v2(i);
    return res;
}

// Vector addition
template <class T, class V1, class V2>
struct vv_add : public v_expr_with_assign<vv_add<T, V1, V2>, T> {

    const v_expr<V1, T> & v1;
    const v_expr<V2, T> & v2;

    vv_add(const v_expr<V1, T> & v1, const v_expr<V2, T> & v2)
      : v1(v1), v2(v2)
    {
    }

    size_t size() const
    {
        return v1.size();
    }

    T operator()(size_t i) const
    {
        return v1(i) + v2(i);
    }
};

template <class T, class V1, class V2>
vv_add<T, V1, V2> operator+(const v_expr<V1, T> & v1,
                            const v_expr<V2, T> & v2)
{
    return vv_add<T, V1, V2>(v1, v2);
}

// Vector subtraction
template <class T, class V1, class V2>
struct vv_sub : public v_expr_with_assign<vv_sub<T, V1, V2>, T> {

    const v_expr<V1, T> & v1;
    const v_expr<V2, T> & v2;

    vv_sub(const v_expr<V1, T> & v1, const v_expr<V2, T> & v2)
      : v1(v1), v2(v2)
    {
    }

    size_t size() const
    {
        return v1.size();
    }

    T operator()(size_t i) const
    {
        return v1(i) - v2(i);
    }
};

template <class T, class V1, class V2>
vv_sub<T, V1, V2> operator-(const v_expr<V1, T> & v1,
                            const v_expr<V2, T> & v2)
{
    return vv_sub<T, V1, V2>(v1, v2);
}

#endif /*MATRIX_VECTOR_H_*/
