#ifndef __MESH_H__
# define __MESH_H__

# include "common.h"
# include "matrix_vector.h"

template <class T> struct mesh_node_2d {

    T x, y;
    std::vector<T> attr;
    int boundary;
    size_t number;

    mesh_node_2d() : x(T()), y(T()), boundary(0), number(0) { }

    bool operator<(const mesh_node_2d & o) const 
    {
        if (y < o.y)
            return false;
        if (o.y < y)
            return true;
        return x < o.x;
    }

};

template <class T> struct mesh_element_2d {

    std::vector<size_t> nodes;
    std::vector<T> attr;

    size_t region() const
    {
        if (0 == attr.size()) 
            return 0;
        else 
            return size_t(attr[0]);
    }
};

template <class T> struct mesh_2d {

    typedef std::vector<mesh_node_2d<T> > node_set;
    typedef std::vector<mesh_element_2d<T> > element_set;
    typedef std::map<size_t, size_t> poly_row;
    typedef std::map<size_t, poly_row> poly_set;

    node_set nodes;
    element_set elements;
    poly_set poly;

    size_t boundary(size_t n1, size_t n2) const
    {
        if (n2 > n1)
            std::swap(n1, n2);
        poly_set::const_iterator i = poly.find(n1);
        if (i == poly.end())
            return 0;
        poly_row::const_iterator j = i->second.find(n2);
        if (j == i->second.end())
            return 0;
        return j->second;
    }

    const mesh_node_2d<T> & node(size_t num) const
    {
        return nodes[num];
    }

    void renumber()
    {
        std::vector<mesh_node_2d<T> > copy = nodes;
        for (size_t N = copy.size(); N--; )
            copy[N].number = N;
        std::sort(copy.begin(), copy.end());
        for (size_t N = copy.size(); N--; ) 
            nodes[copy[N].number].number = N;
    }

    void refine()
    {
        std::vector<mesh_element_2d<T> > new_e;
        std::vector<mesh_node_2d<T> > new_n = nodes;
        size_t N = nodes.size();
        asm_matrix<size_t, true> edges(N);
        foreach (const mesh_element_2d<T> & e, elements) {
            T x[3], y[3];
            size_t num[3], bound[3];
            for (int i = 3; i--;) {
                mesh_node_2d<T> n = node(e.nodes[i]);
                x[i] = n.x; y[i] = n.y; num[i] = e.nodes[i]; 
                bound[i] = n.boundary;
            }

# define mid_node(i)\
            size_t n##i = edges(num[i - 1], num[i%3]);\
            if (!n##i) {\
                n##i = N++;\
                edges(num[i - 1], num[i%3]) = n##i;\
                edges(num[i%3], num[i - 1]) = n##i;\
                mesh_node_2d<T> node;\
                node.x = (x[i - 1] + x[i%3])/2;\
                node.y = (y[i - 1] + y[i%3])/2;\
                node.boundary = boundary(num[i - 1], num[i%3]);\
                new_n.push_back(node);\
            }
            mid_node(1);
            mid_node(2);
            mid_node(3);
# undef mid_node
            mesh_element_2d<T> ne;
            ne.nodes.resize(3);
            ne.attr = e.attr;
            ne.nodes[0] = n1; ne.nodes[1] = n2; ne.nodes[2] = n3;
            new_e.push_back(ne);
            ne.nodes[0] = num[0]; ne.nodes[1] = n1; ne.nodes[2] = n3;
            new_e.push_back(ne);
            ne.nodes[0] = n1; ne.nodes[1] = num[1]; ne.nodes[2] = n2;
            new_e.push_back(ne);
            ne.nodes[0] = n3; ne.nodes[1] = n2; ne.nodes[2] = num[2];
            new_e.push_back(ne);
        }
        poly_set new_p;
        foreach (poly_set::value_type & e1, poly) {
            foreach (poly_row::value_type & e2, e1.second) {
                size_t n1 = e1.first;
                size_t n2 = e2.first;
                size_t m = edges(n1, n2);
                size_t bound = e2.second;
                new_p[m][n1] = bound;
                new_p[m][n2] = bound;
            }
        }
        new_e.swap(elements);
        new_n.swap(nodes);
        new_p.swap(poly);
    }

    void refine(int n) 
    {
        for (int i = n; i--;) {
            refine();
            std::cout << "." << std::flush;
        }
        std::cout << std::endl;
    }

    void load(std::string filename)
    {
//         std::string filetype = filename.substr(filename.find_last_of("."));
//         if (".vol" == filetype) 
//             load_netgen(filename);
//         else if (".msh" == filetype) 
//             load_gmsh(filename);
//         else 
            load_triangle(filename);
    }

    // loads mesh generated with TRIANGLE
    void load_triangle(std::string filename) 
    {
        std::ifstream node_f((filename + ".node").c_str());
        if (!node_f)
            throw std::runtime_error("could not open .node file!");
        size_t N, D, A, B, Offset;
        node_f >> N >> D >> A >> B;
        for (size_t i = N; i--;) {
            mesh_node_2d<T> node;
            size_t num;
            node_f >> num >> node.x >> node.y;
            Offset = num + i + 1 - N;
            for (size_t j = A; j--;) {
                T attr;
                node_f >> attr;
                node.attr.push_back(attr);
            }
            if (B)
                node_f >> node.boundary;
            nodes.push_back(node);
        }
        std::ifstream ele_f((filename + ".ele").c_str());
        if (!ele_f)
            throw std::runtime_error("could not open .ele file!");
        ele_f >> N >> B >> A;
        for (size_t i = N; i--;) {
            mesh_element_2d<T> ele;
            size_t num;
            ele_f >> num;
            for (size_t j = B; j--;) {
                size_t node;
                ele_f >> node;
                ele.nodes.push_back(node - Offset);
            }
            for (size_t j = A; j--;) {
                T attr;
                ele_f >> attr;
                ele.attr.push_back(attr);
            }
            elements.push_back(ele);
        }
        std::ifstream poly_f((filename + ".poly").c_str());
        if (!poly_f)
            throw std::runtime_error("could not open .poly file!");
        poly_f >> A >> A >> A >> A >> N >> B;
        for (size_t i = N; i--;) {
            size_t I, J;
            poly_f >> A >> I >> J >> A;
            if (I < J)
                std::swap(I, J);
            poly[I][J] = A;
        }
    }

};

#endif 
