#ifndef __PCG_MIC0_H__
#define __PCG_MIC0_H__

#include "common.h"
#include "matrix_vector.h"

class mic0_failure {
};

template <class T> class mic0 {

    vector<T> mic0_;
    const sparse_matrix<T, true> & matrix_;

    T neg0(T t) const
    {
        return int(t < T())*t;
    }
    
    void solve_lower(vector<T> & x) const
    {
        size_t n = matrix_.size();
        for (size_t i = 0; i < n; ++i) {
            x(i) /= mic0_(i);
            for (typename sparse_matrix<T, true>::row_iterator
                     it = matrix_.row_begin(i),
                     end = matrix_.row_end(i); ++it != end; )
                x(it.column()) -= neg0(it.value())*x(i);
        }
    }

    void solve_diag(vector<T> & x) const
    {
        for (size_t i = mic0_.size(); i--;)
            x(i) *= mic0_(i);
    }

    void solve_upper(vector<T> & x) const
    {
        size_t n = matrix_.size();
        for (size_t i = n; i--;) {
            for (typename sparse_matrix<T, true>::row_iterator
                     it = matrix_.row_begin(i),
                     end = matrix_.row_end(i); ++it != end; )
                x(i) -= neg0(it.value())*x(it.column());
            x(i) /= mic0_(i);
        }
    }
    
 public:
    explicit mic0(const sparse_matrix<T, true> & m)
      : matrix_(m), mic0_(m.size())
    {
        size_t n = m.size();
        // TODO: this parameter needs adjustment
        T ksi = 5/n;
        for (size_t i = n; i--;) 
            mic0_(i) = m.row_begin(i).value(); // x_i = a_{ii}
        size_t pos_el = 0;
        for (size_t i = 0; i < n; ++i) {
            T w = T();
            for (typename sparse_matrix<T, true>::row_iterator
                     it = m.row_begin(i),
                     end = m.row_end(i); ++it != end; ) {
                T val = it.value();
                if (val > T()) {
                    // diagonal compensation
                    mic0_(i) += val;
                    mic0_(it.column()) += val;
                    ++pos_el;
                } else {
                    w -= it.value();
                }
            }
            // diagonal perturbation
            if (mic0_(i) >= 2*w) 
                mic0_(i) *= (1. + ksi);
            else 
                mic0_(i) *= (1. + sqrt(ksi));
        }
        std::cout << "MIC(0) positive elements: " << pos_el << std::endl;
        for (size_t k = 0; k < n; ++k) {
            T sum = T(); // \sum_{j=k+1}^{n} a_{kj}
            for (typename sparse_matrix<T, true>::row_iterator
                     it = m.row_begin(k),
                     end = m.row_end(k); ++it != end; )
                sum += neg0(it.value());
            for (typename sparse_matrix<T, true>::row_iterator
                     it = m.row_begin(k),
                     end = m.row_end(k); ++it != end; )
                // x_i -= sum a_{ki}/x_k
                mic0_(it.column()) -= sum*neg0(it.value())/mic0_(k); 
        }
        for (size_t i = 0; i < n; ++i) {
            if (mic0_(i) <= T())
                throw mic0_failure();
        }
    }

    void operator()(vector<T> & x) const
    {
        solve_lower(x);
        solve_diag(x);
        solve_upper(x);
    }

    void operator()(vector<T> & x, const vector<T> & b) const
    {
        x = b;
        (*this)(x);
    }
    
};

template <class T> T max_abs(const vector<T> & v)
{
    T max = T();
    for (size_t i = v.size(); i--;) {
        T t = v(i);
        if (t < T())
            t = -t;
        if (t > max)
            max = t;
    }
    return max;
}

template<class T> size_t pcg(vector<T> & x,
                             const sparse_matrix<T, true> & A,
                             const vector<T> & b,
                             T eps)
{
    mic0<T> M(A);
    size_t N = A.size();
    vector<T> r = A*x;
    r = b - r;
    vector<T> z(N);
    M(z, r); // z = M^{-1} r
    vector<T> p = z;
    T rz = (r, z);
    vector<T> Ap(N);
    size_t it = 0;
    for (; max_abs(r) > eps; ++it) {
        Ap = A*p;
        T a = rz/(p, Ap);
        x = x + a*p;
        r = r - a*Ap;
        M(z, r);
        T rz0 = rz;
        rz = (r, z);
        T b = rz/rz0;
        p = z + b*p;
        std::cout << "." << std::flush;
    }
    std::cout << std::endl;
    return it;
}

#endif
