// Calc.h
#ifndef __calc_h_included__
#define __calc_h_included__

#include <iostream>
#include <functional>
#include <string>
#include <cmath>
#include <typeinfo>
#include <map>
#include <vector>
#include <stack>
#include <iomanip>
#include <algorithm>
#include <stdexcept>
#include "poly.h"
#include <cassert>
#include "small.h"

namespace My
{
using namespace std;


namespace Calc {

template<class T>
T Max(const T& X, const T&Y)
{
	return (X>=Y)?X:Y;
}

template<class T>
T Min(const T& X, const T&Y)
{
	return (X<=Y)?X:Y;
}


enum CalcType{BASE, VARIABLE, UNARY_OPERATOR, BINARY_OPERATOR};

class Base{
public:
	virtual CalcType GetType() const
	{
		return BASE;
	}
  virtual ~Base(){}
};


template<class T>
class Variable: public Base
{
	T t;
public:
	Variable(const T&a):t(a){}
	void Set(const T &a)
	{
		t = a;
	}
	T operator()() const
	{
		return t;
	}
	virtual CalcType GetType() const
	{
		return VARIABLE;
	}
};

class Operator:public Base
{
	const int priority;
public:
	Operator(int a):priority(a){}
	int GetPrority(){return priority;}
};

template<class T>
class UnaryOperator:public Operator
{
	typedef T (*fncptr)(const T&);
public:
	UnaryOperator(const int p, const fncptr f):Operator(p), func(f)
	{
	}
	const fncptr func;
	virtual CalcType GetType() const
	{
		return UNARY_OPERATOR;
	}
};

template<class T>
class BinaryOperator:public Operator
{
	typedef T (*fncptr)(const T&, const T&);
public:
	BinaryOperator(const int p, const fncptr f):Operator(p), func(f)
	{
	}
	const fncptr func;
	virtual CalcType GetType() const
	{
		return BINARY_OPERATOR;
	}
};

template<class T>
T Id(const T &X)
{
	return X;
}

template<class T>
T Minus(const T &X)
{
	return -X;
}
template<class T>
T Minus(const T &X, const T &Y)
{
	return X - Y;
}

template<class T>
T Mul(const T &X, const T &Y)
{
	return X*Y;
}
template<class T>
T Div(const T &X, const T &Y)
{
	return X/Y;
}

template<class T>
T Plus(const T &X, const T &Y)
{
	return X + Y;
}

template<class T>
T Pow(const T &X, const T &Y)
{
	return pow(X, Y);
}
template<class T>
T Abs(const T &X)
{
	return (X>=0)?X:-X;
}

template<class T>
T Sin(const T &X)
{
	return sin(X);
}

template<class T>
T Cos(const T &X)
{
	return cos(X);
}
template<class T>
T Tan(const T &X)
{
	return tan(X);
}
template<class T>
T ATan(const T &X)
{
	return atan(X);
}

class ExprTokenizer
{
	string expr;
	int i;
	int l;
public:
	ExprTokenizer(const string &a)
		:expr(a)
	{
		i = 0;
		l = a.length();
	}
	enum TokenType{OPERATOR, CONSTANT, END};
	void SkipSpc()
	{
		while((i < l)&&(expr[i]==' '))
		{
			i++;
		}
	}
	TokenType GetNext(string &a)
	{
		SkipSpc();
		if (i >=l)
			return END;
		if (((expr[i]<='9')&&(expr[i]>='0'))||expr[i]=='.')
		{
			a = "";
			while ((i < l) && (expr[i]<='9') && (expr[i] >= '0'))
			{
				a+=expr[i++];
			}
			if (i >= l)
				return CONSTANT;
			if (expr[i] == '.')
			{
				a+='.';
				i++;
				if (i >= l)
					return CONSTANT;
				while ((i < l) && (expr[i]<='9') && (expr[i] >= '0'))
				{
					a+=expr[i++];
				}
				if (i >= l)
					return CONSTANT;
			}
			if (expr[i] == 'e')
			{
				a+= 'e';
				i++;
				if (i >= l)
					return END;
				if ((expr[i] == '+') && (expr[i] == '-'))
				{
					a += expr[i++];
					if (i >= l)
						return END;
				}
				while ((i < l) && (expr[i]<='9') && (expr[i] >= '0'))
				{
					a+=expr[i++];
				}
			}
			return CONSTANT;
		}
		if (((expr[i]>='a') && (expr[i] <= 'z')) || ((expr[i]>='A') && (expr[i]<='Z')))
		{
			a = "";
			while( (i < l) && (((expr[i]>='a') && (expr[i] <= 'z')) || ((expr[i]>='A') && (expr[i]<='Z'))))
				a+=expr[i++];
			return OPERATOR;
		}
		a = expr[i++];
		return OPERATOR;
	}
};




template <class T>
class Expression
{
	Variable<T> * def_var; 

	multimap<string, Variable<T> *> vars;

	vector<Base*> expr;
	stack<T> registers;

	static map<string, BinaryOperator<T> > BinOps;
	static map<string, UnaryOperator<T> > UnOps;
	static bool initialized;
	static void InitOperators();

public:
	Expression(string s)
	{
		typedef typename map<string, Variable<T>* >::value_type Var;
		typename map<string, BinaryOperator<T> >::iterator bi;
		typename map<string, UnaryOperator<T> >::iterator ui;

		InitOperators();

		ExprTokenizer et(s);

		stack<Operator*> opstack;
		ui = UnOps.find("(");
		opstack.push( &((*ui).second));

		string a;
		ExprTokenizer::TokenType t;
		bool UnaryExpected = true;
		while ((t = et.GetNext(a))!= ExprTokenizer::END)
		{
			if (t == ExprTokenizer::CONSTANT)
			{
				if (!UnaryExpected)
					throw invalid_argument("Invalid expression");
				Variable<T> *ptr;
				vars.insert(Var("Constant", ptr = new Variable<T>(atof(a.c_str()))));
				expr.push_back(ptr);
				UnaryExpected = false;
			}else // In case of operator;
			{
				if (UnaryExpected)
				{
					ui = UnOps.find(a);
					if (ui != UnOps.end())
					{
						opstack.push( &((*ui).second));
					}else
					{
					  typename multimap<string, Variable<T> *>::iterator i;
						Variable<T> *ptr;
						i = vars.find(a);
						if (i != vars.end())
						{
							ptr = (*i).second;
						}else
						{
							vars.insert(Var(a, ptr = new Variable<T>(0)));
						}
						expr.push_back(ptr);
						UnaryExpected = false;
					}
				}else // Binary expected
				{
					bi = BinOps.find(a);
					if (bi != BinOps.end())
					{
						int prio = ((*bi).second).GetPrority();

						while(opstack.top()->GetPrority()>=prio)
						{
							expr.push_back(opstack.top());
							opstack.pop();
						}
						opstack.push( &((*bi).second));
						UnaryExpected = true;
					}
					else if (a == ")")
					{
						while(opstack.top()->GetPrority()>0)
						{
							expr.push_back(opstack.top());
							opstack.pop();
						}
						opstack.pop(); // pop also the "(";
						if (opstack.size()==0) // if it is the last
							throw invalid_argument("Bad expression.");
						UnaryExpected = false;
					}else throw invalid_argument("Invalid expression");
				}
			}
		}
		if (opstack.size()==0)
			throw invalid_argument("Bad expression.");
		while(opstack.top()->GetPrority()>0)
		{
			expr.push_back(opstack.top());
			opstack.pop();
		}
		opstack.pop(); // pop also the "(";
		if (opstack.size()!=0)
			throw invalid_argument("Bad expression.");
		SetDefVar("x");
		SetValue("e", exp(T(1)));
		SetValue("pi", atan(T(1))*4);
	}

	void SetDefVar(const string &a)
	{
		typename multimap<string, Variable<T>* >::iterator i;
		i = vars.find(a);
		if (i!=vars.end())
		{
			def_var = i->second;
		}else
		{
			typedef typename map<string, Variable<T>* >::value_type Var;
			vars.insert(Var(a, def_var = new Variable<T>(0)));
		}

	}
	void SetValue(const string &key, const T & val)
	{
		typename multimap<string, Variable<T>* >::iterator i;
		i = vars.find(key);
		if (i!=vars.end())
		{
			i->second->Set(val);
		}
	}
	T operator ()(const T &a)
	{
		def_var->Set(a);
		return operator()();
	}
	T operator ()();

	~Expression()
	{
		typename multimap<string, Variable<T>* >::iterator i;
		do {
			i = vars.begin();
			if (i != vars.end())
			{
				Variable <T> * ptr = (*i).second;
				vars.erase(i);
				delete ptr;
			}else break;
		}while(true);
	}
};

template< class T>
map<string, BinaryOperator<T> > Expression<T>::BinOps;
template< class T>
map<string, UnaryOperator<T> > Expression<T>::UnOps;
template<class T>
bool Expression<T>::initialized = false;

template<class T>
void Expression<T>::InitOperators()
{
	if (initialized)
		return;
	initialized = true;
	typedef pair<string, BinaryOperator<T> > Pb;
	typedef pair<string, UnaryOperator<T> > Pu;
	typedef BinaryOperator<T> Bo;
	typedef UnaryOperator<T> Uo;
	BinOps.insert(Pb("*", Bo(2000, Mul)));
	BinOps.insert(Pb("/", Bo(2000, Div)));
	BinOps.insert(Pb("+", Bo(1000, Plus)));
	BinOps.insert(Pb("-", Bo(1000, Minus)));
	BinOps.insert(Pb("^", Bo(3000, Pow)));
	UnOps.insert(Pu("-", Uo(2500, Minus)));
	UnOps.insert(Pu("+", Uo(2500, Id)));
	UnOps.insert(Pu("(", Uo(0, Minus)));
	UnOps.insert(Pu("sin", Uo(5000, Sin)));
	UnOps.insert(Pu("cos", Uo(5000, Cos)));
	UnOps.insert(Pu("tan", Uo(5000, Tan)));
	UnOps.insert(Pu("atan", Uo(5000, ATan)));
	UnOps.insert(Pu("abs", Uo(5000, Abs)));
}


template<class T>
T Expression<T>::operator ()()
{
	vector<Base*>::iterator i;

	for (i = expr.begin(); i != expr.end(); i++)
	{
		if ((*i)->GetType()==VARIABLE)
		{
			registers.push(  (*((Variable<T>*)*i))  ());
		}else
		if ((*i)->GetType()==UNARY_OPERATOR)
		{
			registers.top() = (((UnaryOperator<T>*)(*i))->func)(registers.top());
		}else
		{
			// type shuld be == BINARY_OPERATOR;
			T t = registers.top();
			registers.pop();
			registers.top() = (((BinaryOperator<T>*)(*i))->func)(registers.top(), t);

		}
		
	}
	T t = registers.top();
	registers.pop();
	return t;
}












template<class T, class Scalar=T>
class Matrix
{
protected:
	bool Transponated;
public:
	Matrix()
	{
		Transponated = false;
	}
  virtual ~Matrix(){}
	virtual T&operator()(int, int) = 0;
	virtual T operator()(int, int) const = 0;
	virtual int Rows() const = 0;
	virtual int Columns() const = 0;
	virtual void Zero()
	{
		for (int i = 1; i <= Rows(); i++)
			for (int j = 1; j<=Columns(); j++)
				operator()(i, j) = 0;
	}
	void Transponate()
	{
		Transponated = !Transponated;
	}
	virtual void XchangeRow(int f, int s);  // exchanges row f with row s;

	void MulRow(int r, const T&); // multiplys row r with a;
	void MulRowScalar(int r, const Scalar&); // multiplys row r with a;

	void MulRowAdd(int row1, int row2, const T&);// multiplys row2 with a and adds it to row 1
	void MulRowAddScalar(int row1, int row2, const Scalar&);// multiplys row2 with a and adds it to row 1

};

template<class T>
class SomewhereMatrix:public Matrix<T>
{
	const int width;
	const int height;
	T*storage;
public:
	SomewhereMatrix(T*where, int h, int w)
		:storage(where),width(w), height(h)	{}
	SomewhereMatrix(T*where, int i, int h, int w)
		:storage(where+(i-1)*h*w),width(w), height(h)	{}

	int Columns() const 
	{
		return width;
	}

	int Rows()const 
	{
		return height;
	}

	virtual T&operator()(int i, int j)
	{
		return storage[(i-1)*width + j - 1];
	}
	virtual T operator()(int i, int j) const 
	{
		return storage[(i-1)*width + j - 1];
	}

};

template<class T>
class SomewhereMatrix2:public Matrix<T>
{
	int is, js;
	int Width, Height;
	Matrix<T> &a;
	public:
	SomewhereMatrix2(Matrix<T>&mat, int i, int j, int height, int width):a(mat), is(i-1), js(j-1), Width(width), Height(height){}

	int Columns() const 
	{
		return Width;
	}

	int Rows()const 
	{
		return Height;
	}

	virtual T&operator()(int i, int j)
	{
		return a(is+i, js+j);
	}
	virtual T operator()(int i, int j) const 
	{
		return a(is+i, js+j);
	}
};


template<class T, class Scalar>
void Matrix<T, Scalar>::XchangeRow(int f, int s)
{
	for (int i=1;i<=Columns();i++)
	{
		T p;
		p=operator()(f,i);
		operator()(f,i)=operator()(s,i);
		operator()(s,i)=p;
	}
}

template<class T, class Scalar>
void Matrix<T, Scalar>::MulRow(int row, const T&a)
{
	for (int i=1;i<=Columns();i++)operator()(row,i)*=a;
}

template<class T, class Scalar>
void Matrix<T,Scalar>::MulRowScalar(int row, const Scalar&a)
{
	for (int i=1;i<=Columns();i++)operator()(row,i)*=a;
}


template<class T, class Scalar>
void Matrix<T, Scalar>::MulRowAdd(int row1, int row2, const T&a)
{
	for (int k=1;k<=Columns();k++)operator()(row1,k)+=a*operator()(row2,k);
}
		
template<class T, class Scalar>
void Matrix<T,Scalar>::MulRowAddScalar(int row1, int row2, const Scalar&a)
{
	for (int k=1;k<=Columns();k++)operator()(row1,k)+=a*operator()(row2,k);
}




template<class T>
class FullMatrix:public Matrix<T>
{
	const int Row;
	const int Column;
	T *storage;
	using Matrix<T>::Transponated;
public:
	FullMatrix(int i, int j):Row(i), Column(j)
	{
		storage = new T[Row*Column];
	}
	~FullMatrix()
	{
		delete[] storage;
	}
	int Rows() const { 
		if (!Transponated)
			return Row; 
		else
			return Column;
	}
	int Columns() const { 
		if (!Transponated) 
			return Column; 
		else
			return Row;
	}

	T& operator()(int i, int j)
	{
		if (!Transponated)
			return storage[(i-1)*Column + j - 1];
		else
			return storage[(j-1)*Column +i - 1];
	}
	T operator()(int i, int j) const
	{
		if (!Transponated)
			return storage[(i-1)*Column + j - 1];
		else
			return storage[(j-1)*Column + i - 1];
	}

};


template<class T>
class NDiagSymMx:public Matrix<T>{
	T *storage;
	const int N;
	const int Dia;

public:

	int Rows()const {return N;}
	int Columns()const {return N;}

	int Diag()const {return Dia;}


	NDiagSymMx(int n, int diag):N(n), Dia(diag)
	{
		if (diag > n)
			throw invalid_argument("NDiagSymMx. Diags More then rows!");
		int size = N*Dia;// Dia*(Dia-1)/2 izlishni elementa.
		storage = new T[size+1]; // one more element for returning reference to ... This will be the first one
	}
	void Zero()
	{
		for (int i = 0; i <=N*Dia; i++)
			storage[i] = 0;
	}

	~NDiagSymMx()
	{
		delete[] storage;
	}

	T operator()(int i, int j) const
	{
		const int d = Abs(j-i);
		if (d >= Dia)
			return storage[0];

		i = Min(i, j);
		return storage[(i-1)*Dia + d + 1];
	}

	T& operator()(int i, int j) // a(i,j)// from 1)
	{
		const int d = Abs(j-i);
		if (d >= Dia)
			return storage[0];

		i = Min(i, j);

		return storage[(i-1)*Dia + d + 1];
	}
};

template <class T>
class Vector:public Matrix<T>
{
	T *storage;
	int N;
	using Matrix<T>::Transponated;
public:
	
	Vector():N(0), storage(NULL)
	{
	}

	void SetSize(int a)
	{
		delete[]storage;
		N = a;
		storage = new T[N];
	}

	Vector<T>&operator = (const Vector<T> &a)
	{
		delete[]storage;
		N = a.N;
		if(N>0)
		{
			storage = new T[N];
			for (int i = 0; i <N; i++)
				storage[i] = a.storage[i];
		}else
			storage = NULL;
		return *this;
	}

	int Size()const {return N;}
	int Rows()const
	{
		if(!Transponated)
			return 1;
		else
			return N;
	}
	int Columns()const
	{
		if(Transponated)
			return 1;
		else
			return N;
	}
			
	Vector(int size):N(size)
	{
		storage = new T[size];
	}

	Vector(const Vector<T>&a)
	{
		N=a.N;
		if (N>0)
		{
			storage = new T[N];
			for (int i = 0; i <N; i++)
				storage[i] = a.storage[i];
		}
		else
			storage = NULL;
	}
	
	T &operator()(int i)
	{
		assert(!((i > N) || (i <=0)));
		return storage[i-1];
	}
	T operator()(int i) const
	{
		return storage[i-1];
	}

	T &operator()(int i, int j)
	{
		if (!Transponated)
		{
			return storage[j-1];
		}else
		{
			return storage[i-1];
		}
	}
	T operator()(int i, int j) const
	{
		if (!Transponated)
		{
			return storage[j-1];
		}else
		{
			return storage[i-1];
		}
	}

	
	~Vector()
	{
		delete[] storage;
	}

};

template<class T>
void Copy(Matrix<T>&a, const Matrix<T>&b)
{
	for (int i = 1; i <= a. Rows(); i++)
		for (int j = 1; j <= a. Columns(); j++)
			a(i, j) = b(i, j);
}


template<class T>
void GauseSolveNdiagSym(NDiagSymMx<T> & a, Vector<T> &b, Vector<T> &x)
{
	int N = a.Rows();
	if ((N!= b.Size())||(N != x.Size()))
		throw invalid_argument("Invalid arguments to GauseSolveNdiagSym.");
	int Diag = a.Diag();

	int k;
	for (k = 1; k < N; k++)
	{
		const int M =  Min(N, k + Diag - 1);
		for (int i = M; i >= k+1;i--)
		{
			T aa = - a(i, k)/a(k, k);
			const int D = Min(Diag-1, N-i + 1 );
			for (int j = 0; j < D; j++)
				a(i,i+j)+=aa*a(k, i + j);
			b(i)+=aa*b(k);
		}

	}
	for (k = N; k>=1; k--)
	{
		const int D = Min(Diag, N-k+1);
		T s = b(k);
		for (int i = 1; i < D; i++)
		{
			s-=a(k, k+i)*x(k+i);
		}
		x(k) = s/a(k, k);
	}
}

template<class T>
void PrintMatrix(Matrix<T>& a);
template<class T>
void GauseSolve(Matrix<T> & a, Matrix<T> &b, Matrix<T> &x)
{
	int N = a.Rows();
	T mn;
	int k;
	for (k = 1; k < N; k++)
	{
		for (int i = k+1; i <= N;i++)
		{
			mn = -a(i, k)/a(k, k);

			a.MulRowAdd(i, k, mn);
			b.MulRowAdd(i, k, mn);
		}
	}
	for (k = N; k>=1; k--)
	{
		T s = b(k,1);
		for (int i = k+1; i <= N; i++)
		{
			s-=a(k, i)*x(i,1);
		}
		x(k,1) = s/a(k, k);
	}
}


template<class T>
void MultiplyMatrix(Matrix<T>& res, Matrix<T>& a, Matrix<T> &b)
{
	if ( (res.Columns() != b.Columns() ) ||
			(res.Rows()!=a.Rows())||
			(a.Columns()!=b.Rows()) )
		throw invalid_argument("Bad count of rows when multiplying matrixes.");
	for (int i = 1; i <= a.Rows(); i++)
		for (int j = 1; j <= b.Columns(); j++)
		{
			T sum = 0;
			for (int k = 1; k <= b.Rows(); k++)
				sum+=a(i, k)*b(k, j);
			res(i, j) = sum;
		}
}


template<class T>
void SumMatrix(Matrix<T>&res, const Matrix<T>&a, const Matrix<T>&b)
{
	for(int i = 1; i <= res.Rows(); i++)
		for (int j = 1; j <= res.Columns(); j++)
			res(i,j) = a(i,j) + b(i,j);
}

template<class T>
void SubMatrix(Matrix<T>&res, const Matrix<T>&a, const Matrix<T>&b)
{
	for(int i = 1; i <= res.Rows(); i++)
		for (int j = 1; j <= res.Columns(); j++)
			res(i,j) = a(i,j) - b(i,j);
}


template<class T>
void MulMatrixNum(Matrix<T>&res, const Matrix<T>&a, const T&b)
{
	for(int i = 1; i <= res.Rows(); i++)
		for (int j = 1; j <= res.Columns(); j++)
			res(i,j) = a(i,j) * b ;
}


template<class T>
void InitMatrix(Matrix<T>&res, const T&one = 1)
{
	T zero = 0;
	for (int i = 1; i <= res.Rows(); i++)
		for (int j = 1; j <= res.Columns(); j++)
			res(i, j) = (i!=j)? zero : one;
}


template<class T>
void ReverseMatrix(Matrix<T>&res, Matrix<T>&a)
{
	int d = a.Columns();
	res(1,1) = T(1)/a(1,1);
	for (int r = 2; r <= d; r++)
	{
		SomewhereMatrix2<T> v(a, 1, r, r-1, 1);
		SomewhereMatrix2<T> u(a, r, 1, 1, r-1);
		SomewhereMatrix2<T> y(res, 1, r, r-1, 1);
		SomewhereMatrix2<T> x(res, r, 1, 1, r-1);
		FullMatrix<T> xt(1, r-1);
		SomewhereMatrix2<T> B(res, 1, 1, r-1, r-1);
		T &ann = a(r,r);
		T &alpha = res(r,r);
		FullMatrix<T> An1(r-1, r-1);
		FullMatrix<T> one(1,1);

		Copy(An1, B);
		MultiplyMatrix(xt, u, An1); // xt is now u*An1
		MultiplyMatrix(one, xt, v);
		MulMatrixNum(x, xt, alpha = T(1)/(one(1,1)-ann)); // we now have x;
		MultiplyMatrix(y, An1, v);
		MultiplyMatrix(B, y, x);
		SubMatrix(B, An1, B);
		MulMatrixNum(y, y, alpha);
		alpha*=-1;
	} 
}

template<class T>
void ReverseMatrix2(Matrix<T>&res, Matrix<T>&a)
{

	int d;
	d = a.Columns();
	if ((d!=a.Rows())||(d!=res.Rows())||(d!=res.Columns()))
		throw invalid_argument("Reverse2:not square matrix.");;
	Matrix<T> &x = res;
	InitMatrix(x);
	int i,j;

	for (j=1;j<=d;j++)
	{
		T m=0;;
		T f=1/a(j,j);
		a.MulRow(j,f);
		x.MulRow(j,f);
		for (i=1;i<=d;i++)
		{
			if(i!=j)
			{
				f=-a(i,j);
				a.MulRowAdd(i,j,f);
				x.MulRowAdd(i,j,f);
			}
		}
	}	
}

template<class T>
void MySolveNdiag(Matrix<T> &AA, Vector<T> &BB, Vector<T> &XX, const int Diag)
{
	int MatrixSize = Diag;
	int BigSize = (AA.Rows()+MatrixSize -1)/MatrixSize;
	int MatrixS2=MatrixSize*MatrixSize;
	T * Asrc, *Bsrc, *Csrc, *Fsrc, *Xsrc;
	Asrc = new T[MatrixS2*BigSize];
	Bsrc = new T[MatrixS2*BigSize];
	Csrc = new T[MatrixS2*BigSize];
	Fsrc = new T[BigSize*MatrixSize];
	Xsrc = new T[BigSize*MatrixSize];
	int j;
	int k;
	int nn = AA.Rows();
	for (k=1; k <=BigSize; k++)
	{
		for (int i = 1; i <= MatrixSize; i++)
		{
			for (int j = 1; j <=MatrixSize; j++)
			{
				SomewhereMatrix<T> A(Asrc, k, MatrixSize, MatrixSize);
				SomewhereMatrix<T> B(Bsrc, k, MatrixSize, MatrixSize);
				SomewhereMatrix<T> C(Csrc, k, MatrixSize, MatrixSize);
				if (k>1)
					A(i, j) = AA( (k-1)*MatrixSize + i, (k-2)*MatrixSize + j);
				else
					A(i,j)= 0;

				if ((k-1)*MatrixSize + i <=nn)
					B(i,j) = AA( (k-1)*MatrixSize + i, (k-1)*MatrixSize + j);
				else
					B(i,j) = (j==i);

				if (((k)*MatrixSize + j <=nn))
					C(i,j) = AA( (k-1)*MatrixSize + i, (k-0)*MatrixSize + j);
				else
					C(i,j) = 0;
			}
			SomewhereMatrix<T> F(Fsrc, k, MatrixSize, 1);
			if((k-1)*MatrixSize+i<=nn)
				F(i, 1) = BB((k-1)*MatrixSize + i);
			else
				F(i, 1) = 0;
		}
	}

	// now we have the system ready...

	int First = 1;
	int Last = BigSize;
	int step = 1;
	stack<int> st;

	// some temporary matrices
	FullMatrix<T> Brev(MatrixSize, MatrixSize);
	FullMatrix<T> temp1(MatrixSize, MatrixSize);
	FullMatrix<T> temp2(MatrixSize, 1);
	FullMatrix<T> CB(MatrixSize, MatrixSize);
	FullMatrix<T> AB(MatrixSize, MatrixSize);


	while(First<Last)
	{
		st.push(Last);
		int i;
		for (i = First; i<= Last;i+=step*2)
		{
			int im1 = i-step;
			int ip1 = i+step;
			SomewhereMatrix<T> Am1(Asrc + (im1-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> Bm1(Bsrc + (im1-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> Cm1(Csrc + (im1-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> Ap1(Asrc + (ip1-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> Bp1(Bsrc + (ip1-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> Cp1(Csrc + (ip1-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> A(Asrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> B(Bsrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> C(Csrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> Fm1(Fsrc + (im1-1)*MatrixSize, MatrixSize, 1);
			SomewhereMatrix<T> Fp1(Fsrc + (ip1-1)*MatrixSize, MatrixSize, 1);
			SomewhereMatrix<T> F(Fsrc + (i-1)*MatrixSize, MatrixSize, 1);

			Copy(temp1, B);
			ReverseMatrix(Brev, temp1);


			if (im1>=First)
			{
				MultiplyMatrix(CB, Cm1, Brev);
				MultiplyMatrix(temp1, CB, A);
				SubMatrix(Bm1, Bm1, temp1);

				MultiplyMatrix(temp2, CB, F);
				SubMatrix(Fm1, Fm1, temp2);

				MultiplyMatrix(Cm1, CB, C);
				MulMatrixNum(Cm1, Cm1, T(-1));
			}
			if (ip1<=Last)
			{
				MultiplyMatrix(AB, Ap1, Brev);
				MultiplyMatrix(temp1, AB, C);
				SubMatrix(Bp1, Bp1, temp1);

				MultiplyMatrix(temp2, AB, F);
				SubMatrix(Fp1, Fp1, temp2);


				MultiplyMatrix(Ap1, AB, A);			
				MulMatrixNum(Ap1, Ap1, T(-1));
			}		
		}
		i-=step*2;
		if (i==Last)
			Last-=step;

		First+=step;
		step*=2;
	}
	// Reverse way.

	int i;
	i = First;
	SomewhereMatrix<T> B(Bsrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
	SomewhereMatrix<T> F(Fsrc + (i-1)*MatrixSize, MatrixSize, 1);
	SomewhereMatrix<T> X(Xsrc + (i-1)*MatrixSize, MatrixSize, 1);
//#define ReverseSolving
#ifdef ReverseSolving
	Calc::ReverseMatrix(temp1, B);
	MultiplyMatrix(X, temp1, F);
#else
	GauseSolve(B, F, X);
#endif


	while(!st.empty())
	{
		step/=2;
		First-=step;
		Last = st.top();
		st.pop();
		for (int i = First; i <=Last; i+=step*2)
		{
			int im1 = i-step;
			int ip1 = i+step;
			SomewhereMatrix<T> A(Asrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> B(Bsrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> C(Csrc + (i-1)*MatrixS2, MatrixSize, MatrixSize);
			SomewhereMatrix<T> F(Fsrc + (i-1)*MatrixSize, MatrixSize, 1);
			SomewhereMatrix<T> X(Xsrc + (i-1)*MatrixSize, MatrixSize, 1);
			SomewhereMatrix<T> Xm1(Xsrc + (im1-1)*MatrixSize, MatrixSize, 1);
			SomewhereMatrix<T> Xp1(Xsrc + (ip1-1)*MatrixSize, MatrixSize, 1);

			if (im1>=First)
			{
				MultiplyMatrix(temp2, A, Xm1);
				SubMatrix(F, F, temp2);
			}
			if (ip1<=Last)
			{
				MultiplyMatrix(temp2, C, Xp1);
				SubMatrix(F, F, temp2);
			}
#ifdef ReverseSolving
			Calc::ReverseMatrix(temp1, B);
			MultiplyMatrix(X, temp1, F);
#else
			GauseSolve(B, F, X);
#endif

		}
	}

	for (j=1; j <=AA.Rows(); j++)
		XX(j) = Xsrc[j-1];

	delete[] Asrc;
	delete[] Bsrc;
	delete[] Csrc;
	delete[] Fsrc;
	delete[] Xsrc;
}

template<class T>
void PrintMatrix(Matrix<T>& a)
{
	cout.precision(4);
	for (int i = 1; i <= a.Rows(); i++)
	{
		for (int j = 1; j<= a.Columns(); j++)
		{
			cout << setw(10) << a(i, j) ;
		}
		cout << '\n';
	}
	cout << endl;
}

template<class T>
class Polinom {
  int Step;
  T *storage;
public:
  
  Polinom(const T t= 0){
    storage = new T[1];
    storage[0] = t;
    Step = 0;
  }
  Polinom(int step):Step(step)
	{
		storage = new T[Step + 1];
		for (int i = 0; i <= Step; i++)
			storage[i] = 0;
	}
	Polinom(const Polinom<T> &a):Step(a.Step)
	{
		storage = new T[Step + 1];
		for (int i = 0; i <=Step; i++)
			storage[i] = a.storage[i];
	}
	Polinom<T>& operator = (const Polinom<T> & a)
	{
		if (storage)
			delete[]storage;
		Step = a.Step;
		storage = new T[Step + 1];
		for(int i = 0; i <= Step; i++)
			storage[i] = a.storage[i];
		return *this;
	}

	Polinom<T>& operator *= (const T&a)
	{
		for (int i = 0; i <= Step; i++)
			storage[i]*=a;
		return *this;
	}

	Polinom<T>& operator += (Polinom<T> const &a)
	{
		Polinom<T> tmp = *this + a;
		*this = tmp;
		return *this;
	}
	Polinom<T>& operator *= (Polinom<T> const &a)
	{
		Polinom<T> tmp = *this * a;
		*this = tmp;
		return *this;
	}

	Polinom<T> operator * (const T&a)const
	{
		Polinom<T> c(Step);
		for (int i = 0; i <= Step; i++)
			c.storage[i] = storage[i]*a;
		return c;
	}

	Polinom<T> integrate() const
	{
		Polinom<T> c(Step+1);
		for(int i = 0; i <=Step; i++)
			c.storage[i+1] = storage[i]/(i+1);
		return c;
	}

	Polinom<T> difference() const
	{
		int newStep = (Step>0)?Step - 1:0;
		Polinom<T> c(newStep);
		for (int i = 1; i<=Step; i++)
			c.storage[i-1] = i*storage[i];
		return c;
	}


	Polinom<T> operator * (const Polinom<T> &b) const
	{
		const Polinom<T> & a = *this;
		Polinom<T> c(a.Step + b.Step);
		for(int i = 0; i <= a.Step; i++)
			for (int j = 0; j <= b.Step; j++)
				c.storage[i+j] += a.storage[i]*b.storage[j];
		return c;
	}

	Polinom<T> operator + (const Polinom<T> &b) const
	{
		if (Step>=b.Step)
		{
			Polinom<T> r(Step);
			for (int i = 0; i <= b.Step; i++)
				r.storage[i] = storage[i] + b.storage[i];
			for (int j = b.Step+1; j<=Step; j++)
				r.storage[j] = storage[j];
			return r;
		}else
		{
			Polinom<T> r(b.Step);
			for (int i = 0; i <= Step; i++)
				r.storage[i] = storage[i] + b.storage[i];
			for (int j = Step+1; j<=b.Step; j++)
				r.storage[j] = b.storage[j];
			return r;
		}
	}


  	Polinom<T> operator - (const Polinom<T> &b) const
	{
	  Polinom<T> c = b;
	  c *= -1;
	  return *this + c;
 	}


	T operator()(const T& x)const 
	{
		T r(storage[Step]);
		for (int i = Step - 1; i>=0; i--)
		{
			r = r*x + storage[i];
		}
		return r;
	}
	T& operator[](const int i)
	{
		return storage[i];
	}
	T operator[](const int i)const
	{
		return storage[i];
	}

	int Degree()const
	{
		return Step;
	}

	~Polinom()
	{
		if (storage)
			delete[] storage;
	}
};

template<class T>
T Integrate(const Polinom<T>&p, const T&a, const T&b)
{
	Polinom<T> I = p.integrate();
	return I(b) - I(a);
}

template<class T>
T DotProduct(const Polinom<T> &a, const Polinom<T>&b, const T&left, const T&right)
{
	Polinom<T> I = (a*b).integrate();
	return I(left) - I(right);
}

template<class T>
Polinom<T> OrthoPolinom(int step, const T& left, const T&right)
{
	Polinom<T> x(1);
	x[1] = 1;
	Polinom<T> p2(0);
	p2[0] = 1;
	T gama = - DotProduct(x, p2, left, right)/DotProduct(p2, p2, left, right);
	Polinom<T> p3(1);
	p3[1] = 1;
	p3[0] = gama;
	if (step == 0) 
		return p2;
	if (step == 1)
		return p3;
	for(int i = 1; i< step; i++)
	{
		Polinom<T> p1 = p2;
		p2 = p3;
		T alpha = - DotProduct(x*p2, p2, left, right)/DotProduct(p2, p2, left, right);
		T beta = - DotProduct(x*p2, p1, left, right)/DotProduct(p1, p1, left, right);
		Polinom<T> hh(1);
		hh[1] = 1;
		hh[0] = alpha;
		p3 = hh*p2 + p1*beta;

	}
	return p3;
}

template<class T>
void OrthoRoots(Vector<T> & a, const T&left, const T&right)
{
	int n = a.Size();
	Polinom<long double> o = OrthoPolinom(n, (long double)left, (long double)right);
	PolinomTemplate<long double> p(n);
	int i;
	for (i = 0; i <=n; i++)
		p[i] = o[i];
	vector<long double> x(n);
	vector<long double > r(n);
	int res = -1200;

	if (n>1)
	for (i = 0; i < n; i++)
	{
	  x[i] = i/(long double)(n-1)-(.5l);
	}else
		x[0] = 0;

	long double eps = T(0);
	r = FindRoots(p, x, 0, eps, res, true);

	for (i = 0; i < n; i++)
	  {
		a(i+1) = r[i];
	  }
}

template<class T>
void GetKwadratureCoeff(Vector<T> &coeff, const Vector<T> &points, const T&left, const T&right)
{
	assert(coeff.Size()!=points.Size());
	for (int i = 1; i <=coeff.Size(); i++)
	{
		Polinom<T> p(0);
		p[0] = 1;
		for (int j = 1; j <=coeff.Size(); j++)
			if (i!=j)
			{
				Polinom<T> hh(1);
				hh[1] = 1;
				hh[0] = -points(j);
				p = p*hh;
			}
		p = p*p;
		coeff(i) = T(1)/p(points(i)) * Integrate(p, left, right);
	}

}

template<class T, class D>
void Discretize(Vector<T>& r, D&d, const Vector<T>&points)
{
	for (int i = 1; i <= points.Size(); i++)
		r(i) = d(points(i));
}


template<class T, int n>
SmallMatrix<T, n-1, n-1> Adj(const SmallMatrix<T, n, n> &A, int I, int J)
{
  SmallMatrix<T, n-1, n-1> ret;

  for (int i = 0; i < n; ++i)
  {
    if (i == I)
      continue;
    for (int j = 0; j < n; ++j)
    {
      if (j == J)
	continue;
      int ii = i;
      int jj = j;
      ii -= i>I;
      jj -= j>J;
      ret[ii][jj] = A[i][j];
    }
  }
  return ret;
}
template<class T>
T Deter (SmallMatrix<T, 1, 1>const &A)
{
  return A[0][0];
}


template<class T, int n>
T Deter (SmallMatrix<T, n, n>const &A)
{
  T sign = 1.;
  T res = 0.;
  for (int j = 0; j < n; ++j)
  {
    res += sign * A[0][j] * Deter(Adj(A, 0,j));
    sign *= -1.;
  }
  return res;
}


template<class T, int n>
My::Calc::Polinom<T> CharPolinom(const SmallMatrix<T, n, n> &a)
{
  SmallMatrix<My::Calc::Polinom<T>, n, n> M;
  for (int i = 0; i < n; ++i)
    for (int j = 0; j < n; ++j)
      if (i!=j)
	M[i][j][0] = a[i][j];
      else
      {
	My::Calc::Polinom<T> r(1);
	r[0] = a[i][j];
	r[1] = -1;
	M[i][j] = r;
      }
  return Deter(M);       
}



template<class T, int N>
void SwapRow(SmallMatrix<T, N, N> &A, int I, int J)
{
  if (I != J)
    for (int i = 0; i < N; ++i)
      std::swap(A[I][i], A[J][i]);
}
template<class T, int N>
void SwapCol(SmallMatrix<T, N, N> &A, int I, int J)
{
  if (I != J)
    for (int i = 0; i < N; ++i)
      std::swap(A[i][I], A[i][J]);
}

template<class T, int N>
int MaxElementRow(SmallMatrix<T, N, N> const &A, int col)
{
  int row = col;
  T max = std::abs(A[row][col]);
  T m;
  for (int i = 0; i <col; ++i)
    if (max > (m=std::abs(A[i][col])))
    {
      max = m;
      row = i;
    }
  return row;
}

template<class T, int N>
void DoColumn(SmallMatrix<T, N, N> &A, T factor,  int j, int l)
{
  for (int k = 0; k < N; ++k)
  {
    A[k][l] += factor * A[k][j];
  }
}
    

template<class T, int N>
T Max(SmallMatrix<T, N, N>const &A)
{
  T Max = 0;
  for (int i = 0; i < N; ++i)
    for (int j = 0; j < N; ++j)
    {
      T val = std::fabs(A[i][j]);
      if (val > Max)
	Max = val;
    }
  return Max;
}


template<class T, int N>
void NullSpace(SmallMatrix<T, N, N>const &B, std::vector<SmallVector<T, N> > &ns)
{
  SmallMatrix<T, N, N> mat = B;
  SmallMatrix<T, N, N> I;
  for (int i = 0; i < N; ++i)
    I[i][i] = 1;
  const T maxthres=1e-12;
  for (int j = N; j--;)
  {
    int mr = MaxElementRow(mat, j);
    SwapRow(mat, j, mr);
    if (fabs(mat[j][j])<maxthres)
    {
      std::cout << j << " " << mat [j][j] << std::endl;
      
      std::cout << "Zero ! " << std::endl;
      std::cout << " Mat - " << mat << std::endl;
      continue;
    }
    for (int l = 0; l < j; ++l)
      if (l!=j)
      {
	T factor = -mat[j][l]/mat[j][j];
	DoColumn(mat, factor, j, l);
	DoColumn(I, factor, j, l);
      }
    T factor = mat[j][j];
    for (int i = 0; i < N; ++i)
    {
      // mat[i][j]/= factor;
      // I[i][j]/= factor;
    }
  }
  // check for zero
  //  std::cout.precision(20);
  T zerothres = Max(mat)*maxthres;
  std::cout << " Mat = " << mat << std::endl;
  for (int j = 0; j < N; ++j)
  {
    T sum = T();
    bool flag = true;
    for (int i = 0; i < N; ++i)
      if (fabs(mat[i][j])>maxthres)
	flag = false;
    if (flag)
    {
      SmallVector<T, N> v;
      T sum = T();
      for (int i = 0; i < 3; ++i)
      {
	v[i] = I[i][j];
	sum += v[i]*v[i];
      }
      v/= sqrt(sum);
      ns.push_back(v);
    }
  }
}

template<class T, int N>
void REigenStuff(SmallMatrix<T, N, N> const &A, SmallVector<T, N> &lambda, SmallMatrix<T, N, N> &v)
{
  
  typedef std::complex<T> CT;
  My::Calc::Polinom<T> pol = My::Calc::CharPolinom(A);
  PolinomTemplate<CT> p(N);
  for (int i = 0; i <=N; ++i)
    p[i] = pol[i];
  int res = -12000;
  long double eps = 0;

  std::vector<CT> roots(N);
  for (int i = 0; i < N; i++)
  {
    const double PI = atan(1.)*4;
    double angle = 2.*PI*i/N+1.2;
    double r = std::cos(angle);
    double imag = std::sin(angle);
    roots[i] = CT(r, imag);
  }

  roots = FindRoots(p, roots, 0, eps, res, true);
  for (int i = 0; i < N; ++i)
    lambda[i] = std::real(roots[i]);
  std::sort(&lambda[0], &lambda[0]+3, std::greater<double>());
  std::vector<SmallVector<T, N> > ns;
  for (int i = 0; i < N; ++i)
  {
    SmallMatrix<T, N, N> mat = A;
    const T l = lambda[i];
    for (int k = 0; k < N; ++k)
      mat[k][k] -= l;
    std::cout << "l = " << l << std::endl;
    NullSpace(mat, ns);
    std::cout << "Size " << i << " "<< ns.size() << std::endl;
    if (ns.size()>=N)
      break;
  }
  assert(ns.size()==N);
  v = T();
  for (int i = 0; i < ns.size(); i++)
    std::cout << "V[" << i << "] = " << ns[i] <<std::endl;
  for (int i = 0; i < N; ++i)
    for (int j = 0; j < ns.size(); ++j)
      v[i][j] = ns[j][i];
  
}




} // namespace calc
} // namespace My

#endif
