#ifndef __cg_hpp__
#define __cg_hpp__

#include "cg.h"
#include <iostream>
#include <iomanip>
#include <sstream>

#include "functors.h"
#include "visitors.h"
#include "hypre_precon.h"


template<class Double, class Int>
NewWorkerScalar<Double, Int>::NewWorkerScalar(MPI::Intracomm  & c,
			    Int n1, Int n2, Int n3)
  :  N1(n1), N2(n2), N3R(n3),comm(c.Dup()), 
     rank(c.Get_rank()), size(c.Get_size())
{
  size = comm.Get_size();
  rank = comm.Get_rank();
  StartLine = rank*(N3R+1)/size;
  Lines = (rank+1)*(N3R+1)/size - StartLine;

  mat.mat.SetSize(N1, N2, Lines, N3R);
  mat.mat.SetComm(comm);
  rhs.SetSize(N1, N2, Lines, N3R);
  rhs.SetComm(comm);
  x.SetSize(N1, N2, Lines, N3R);
  x.SetComm(comm);
}

#include "solver.h"
#include "precon.h"
template<class Double, class Int>
Double NewWorkerScalar<Double, Int>::CheckSolution()
{
  Index3dBase<Int> ind(N1, N2, Lines);
  SmallVector<Double, 3> h,p,  pl, pd, pf, d,d1;

  typedef matrix_type Matrix;
  typedef bvector_type XType;
  typedef bvector_type RHSType;
  typedef typename XType::BType DType;
  typedef typename XType::BType GType;
  typedef typename XType::BType HdType;
  


  h[0] = Options::GetOptions().DM1/ N1;
  h[1] = Options::GetOptions().DM2/ N2;
  h[2] = Options::GetOptions().DM3/ N3R;
  Double maxerr = 0;
  Double maxerr1 = 0;
  for(int i = 0; i < N1; i++)
    for(int j = 0; j < N2; j++)
      for(int k = 0; k < Lines-1; k++)
      {
	ind.SetPos(i, j, k);
	p[0] = (Double)i *h[0];
	p[1] = (Double)j *h[1];
	p[2] = (Double)(k + StartLine) *h[2];
	pl = p;
	pd = p;
	pf = p;
	pl[1] += h[1]*0.5;
	pl[2] += h[2]*0.5;
	
	pf[0] += h[0]*0.5;
	pf[2] += h[2]*0.5;
	  
	pd[0] += h[0]*0.5;
	pd[1] += h[1]*0.5;
	  
	d[0] = U(pl) - x[ind][LeftEl];
	d[1] = U(pf) - x[ind][FrontEl];
	d[2] = U(pd) - x[ind][DownEl];
	
	for(int f = 0; f < 3; ++f)
	{
	  if (fabs(d[f])>maxerr) 
	    maxerr = fabs(d[f]); 
	}
	
      }
  comm.Allreduce(&maxerr, &maxerr1, 1, Myf<Double>::Datatype(), MPI::MAX);
  return maxerr1;
}

template<class Double, class Int>
int
NewWorkerScalar<Double, Int>::
Solve(Double eps, SolutionMethod method, DiscretizationType distype, BoundaryConditions bc)
{
  const int fcs=30;
  const int scs=30;
  typedef PCGSolverBase<bvector_type> sb_type;
  std::auto_ptr<sb_type > solv (
				Options::GetOptions().gcg? (sb_type *)new GPCGSolver<bvector_type> : (sb_type *)new PCGSolver<bvector_type>);
  sb_type  &solver = *solv;
  solver.SetMatrix(&mat);
  solver.SetEps(eps);
  solver.SetSizes(N1, N2, Lines, N3R);
  solver.SetComm(comm);
  ClearVector(x,1);
  PreconBase<bvector_type> *precon = NULL;
  switch(method)
  {
  case smCG:
    precon = new NoPrecon<bvector_type>();
    break;
  case smPCG1:
    {
      MIC0Precon<matrix_type, bvector_type, bvector_type,  PreconDescriptorM1> *p
	= new MIC0Precon<matrix_type, bvector_type, bvector_type, PreconDescriptorM1>;
      p->Create(mat.mat);
      precon = p;
    }    
    break;
  case smPCG2:
    {
      MIC0Precon<matrix_type, bvector_type, bvector_type, PreconDescriptorM2> *p
	= new MIC0Precon<matrix_type, bvector_type, bvector_type, PreconDescriptorM2>;
      p->Create(mat.mat);
      precon = p;
    }    
  
    break;
  case smPCG3:
      {
      MIC0Precon<matrix_type, bvector_type, bvector_type,  PreconDescriptorM3> *p
	= new MIC0Precon<matrix_type, bvector_type, bvector_type,  PreconDescriptorM3>;
      p->Create(mat.mat);
      precon = p;
    }    

    break;
  case smPCG4:
    {
      MIC0PreconM4<matrix_type, bvector_type, bvector_type, PreconDescriptorM4> *p
	= new MIC0PreconM4<matrix_type, bvector_type, bvector_type,  PreconDescriptorM4>;
      p->Create(mat.mat);
      precon = p;
    }    
    break;
  case smBoomer:
    {

      Int NumberOfLines = Lines+1;
      Int SL = StartLine-1;
      Int SO = -1;
      if (rank == 0)
      {
	SL++;
	SO++;
	NumberOfLines--;
      }
      if (rank +1 == size)
      {
	NumberOfLines--;
      }
      
      HyprePrecon<bvector_type>* hp = NULL;      
      hp = new HyprePrecon<bvector_type>;
      hp->Init(comm, N1, N2, N3R);
      DiscretizationType dt=distype;//RP;
      Discretizator<double> dis(dt);
      dis.GenerateMatrix(hp->mat, hp->hdst, NumberOfLines, SL, SO, N3R, bc);
      hp->Init2();
      precon = hp;
    }    
    break;
  default:;
  }
  solver.SetPrecon(precon);



  if (rank == 0)
  {
    cout << "Starting  on " 
	 << size << " processors. Eps is "<< eps<< endl;
    
    cout << setw(fcs) << left << "Problem size "
	 << N1 << "x" << N2 << "x" << N3R << " " << Myf<Double>::ToString() << " "
	 << algname + solver.GetName() <<endl;
    cout << setw(fcs) << left  << "Running on "  
	 << setw(scs) <<size  << "procesors." << endl;
  }
  int iterations  = solver.Solve(x, rhs);
  if (rank == 0)
  {
    cout << setw(fcs) << "Iterations" 
	 << setw(scs) << iterations << endl;
  }
  delete precon;
  return iterations;
}


template<class Double, class Int>
void
NewWorkerScalar<Double, Int>::
Sysgen(DiscretizationType distype, BoundaryConditions bc)
{
  switch (distype)
  {
  case MV: algname+= "MV";break;
  case MP: algname+= "MP";break;
  case RP: algname+= "RP";break;
  case RV: algname+= "RV";break;
  }
  algname += " ";
  if(bc & bcLeft)
    algname += "l";
  if(bc & bcRight)
    algname += "r";
  if(bc & bcDown)
    algname += "d";
  if(bc & bcUp)
    algname += "u";
  if(bc & bcFront)
    algname += "f";
  if(bc & bcBack)
    algname += "b";
  algname += " ";

  Int NumberOfLines = Lines+1;
  Int SL = StartLine-1;
  Int SO = -1;
  if (rank == 0)
  {
    SL++;
    SO++;
    NumberOfLines--;
  }
  if (rank +1 == size)
  {
    NumberOfLines--;
  }
  coefficientvec.SetSize(N1, N2, NumberOfLines);
  std::ostringstream os;
  os << N3R << ".bone";
  coefficientvec.SetPars(Options::GetOptions().zero, 0, 1, 0, 1, (double)SL/N3R, (double)(NumberOfLines+SL)/N3R);
  for(int i = 0; i < N1; ++i)
    for(int j = 0; j < N2; ++j)
      for(int k = 0; k < NumberOfLines; ++k)
	coefficientvec(i,j,k) = 1.;
  if(Options::GetOptions().usebone)
    try{
      coefficientvec.Read(os.str(), SL);//, NumberOfLines);
    }catch(std::exception&)
    {
      std::cerr << "could not open file" << os.str() << std::endl;
    }
  {
    ClearMatrix(mat.mat);
    ClearVector(rhs);
    Discretizator<Double> dis(distype);
    dis.GenerateMatrix(mat.mat, rhs, NumberOfLines, SL, SO, N3R, bc);
    //	  ClearVector(rhs, 1);
  }
}



#endif

