#ifndef __elastsolver_hpp__
#define __elastsolver_hpp__
#include "elastsolver.h"
#include "precon.h"
#ifdef MY_HYPRE
  #include "hypre_precon.h"
#endif
#include "useful.h"

#include "functors.h"

#include <string>
#include <iostream>
#include <sstream>
#ifdef _OPENMP
#include <omp.h>
#endif

#include <memory>

template<class Double, class Int, class BBvec>
ElastWorker<Double, Int, BBvec>:: ~ElastWorker()
{
  delete mat;
}

template<class Double, class Int, class BBvec>
ElastWorker<Double, Int, BBvec>::ElastWorker(MPI::Intracomm  & c)
  :  comm(c),
     N1(Options::GetOptions().N1),
     N2(Options::GetOptions().N2),
     N3R(Options::GetOptions().N3),
     rank(c.Get_rank()), size(c.Get_size())
{
  mat = NULL;
  distype = Options::GetOptions().distype;
  if((N1&1)||(N2&1)||(N3R&1))
    throw std::string("Sizes must be even");

  StartLine = rank*(N3R+1)/size;
  Lines = (rank+1)*(N3R+1)/size - StartLine;
  
  FirstFirst = rank==0;
  LastLast = rank == (size-1);
  FirstFull = ((StartLine &1) == 0);
  
  DoubleLines = (Lines&1)? (Lines+1)/2-Int(FirstFull) : (Lines)/2;
  //  cout << "rank " << rank << " " << Lines << " " << (Int)FirstFull << " "<< DoubleLines << endl;
  //  exit(0);

  if (!Options::GetOptions().reuse_rhs)
  {
    dissol.SetSize(N1, N2, Lines, N3R);
    dissol.SetComm(comm);
    dissol.CreateDatatypes();
  }
  rhs.SetSize(N1, N2, Lines, N3R);
  rhs.SetComm(comm);
  rhs.CreateDatatypes();
  x.SetSize(N1, N2, Lines, N3R);
  x.SetComm(comm);
  x.CreateDatatypes();
  //    cout << "Lines " << rank << " " << Lines << endl;
}

template<class Double, class Int, class BBvec>
void
ElastWorker<Double, Int, BBvec>::
Sysgen()
{

  bc = Options::GetOptions().bc;
  mat =  Options::GetOptions().PBC? new ElastMatrixPBC<BBvec>(comm) : new ElastMatrix<BBvec>(comm);
  mat->Sysgen(rhs,  ( !Options::GetOptions().reuse_rhs) ?dissol:x);

}

template<class Double, class Int, class BBvec>
void
ElastWorker<Double, Int, BBvec>::
WriteSolution()
{
  static bool firsttime = true;
  static int itop[6][8] = {{1,2,3,4,17,18,19,20},
			   {17,18,19,20,33,34,35,36},
			   {5,6,10,11,21,22,26,27},
			   {10,11,15,16,26,27,31,32},
			   {7,8,12,13,23,24,28,29},
			   {8,9,13,14,24,25,29,30}};
  //  static int cubes[BSize][4];

  if(firsttime)
  {
    firsttime = false;

    for(int i = 0; i < 6; ++i)
      for(int j = 0; j < 8; ++j)
	--itop[i][j];

  }

  std::ostringstream oss;
  oss << Options::GetOptions().scratchpath
      <<"/dss." //displacements+stresses+strains;
      << Options::GetOptions().outputname 
      << "." << size
      << "." << rank
      << "." << Options::GetOptions().homk
      << "." << Options::GetOptions().homl;

  std::string fname = oss.str();
  ofstream fout(fname.c_str());
  fout << "#"; // comment
  fout << N1 << "\t" << N2 << "\t" << N3R << "\t"
       <<(((distype==MP)||(distype==RP))?0:1) <<"\t"
       << size << "\t" << rank      << "\t"
       << Options::GetOptions().DM1 << "\t"
       << Options::GetOptions().DM2 << "\t"
       << Options::GetOptions().DM3 
       << "\n";
  fout << "#N1<tab>N2<tab>N3<tab>DT<tab>Nproc<tab>Rank<tab>DM1<tab>DM2<tab>DM3\n" ;
  fout.precision(Options::GetOptions().outputprecision);
  SmallVector<Double, 3> displacements;
  SmallMatrix<Double, 3, 3> stresses;
  SmallMatrix<Double, 3, 3> strains;
  SmallVector<Double, 18> small_x;
  typedef typename ElastDiscretizator<Double>::SparceMacroMatrix_type smt;
  typedef typename smt::vector_type small_vector_type;
  small_vector_type medium_x;
  for(Int k = 0; k < DoubleLines*2; k++)
    for(Int i = 0; i < N1; i++)
      for(Int j = 0; j < N2; j++)
      {
	{
	  Index3dBase<Int> ind(N1, N2, Lines);
	  ind.SetPos(i &~1, j&~1, (k&~1) - 1 +Int(FirstFull));
	  Double lambda = mat->elmentw.GetAt(i/2, j/2, k/2).lambda;
	  Double mu = mat->elmentw.GetAt(i/2, j/2, k/2).mu;
	  
	  BigToSmallCpy<bvector_type, small_vector_type> b2scopy1(x, medium_x);
	  DoForMacroElement(b2scopy1, ind);
	  int cube = (k&1) + (j&1)*2 + (i&1)*4;
	  for(int i = 0; i < 6; ++i)
	  {
	    small_x[0  + i] = medium_x[0 + itop[i][cube]];
	    small_x[6  + i] = medium_x[36 + itop[i][cube]];
	    small_x[12 + i] = medium_x[72 + itop[i][cube]];
	  }
	  fout << lambda << "\t";
	  fout << mu << "\t";
	  for(int i = 0; i < 18; ++i)
	    fout << small_x[i] << "\t";
	    
	  fout << "\n";

	}
      }
  
}

template<class Double, class Int,class BBvec>
void
ElastWorker<Double, Int, BBvec>::
CheckSolution()
{
  x.StartSendRecv();
  x.FinishSendRecv();
  if (!Options::GetOptions().reuse_rhs)
    VecVecOper<VecSub>(dissol, x);      // G -= RHS;

  serr =  VecReduceNorm<CNorm>((!Options::GetOptions().reuse_rhs)?dissol:x);
  if (!Options::GetOptions().reuse_rhs)
  {
    dissol.CreateDatatypes();
    mat->Multiply(dissol, x);
    rhs.AddBoundarys(2);
    dissol.AddBoundarys(2);
    VecVecOper<VecSub>(dissol, rhs);
    //    std::cout << "smqtame " << std::endl;
//     dissol.SetPeriodic(true);
//     rhs.SetPeriodic(true);
    merr =  VecReduceNorm<CNorm>(dissol);
  }else
  {
    merr =  VecReduceNorm<CNorm>(rhs);
  }
};


template<class Double, class Int, class BBvec>
int
ElastWorker<Double, Int, BBvec>::
Solve()
{
  Double eps = Options::GetOptions().eps;
  SolutionMethod method = Options::GetOptions().method;
  
  typedef SmallVector<Double, 6> row_type;
  typedef BDom<row_type, Int> matrix_type;
  typedef BDom<Double, Int> plainvector_type;
  typedef SmallVector<float, 6> srow_type;
  typedef BDom<srow_type, Int> smatrix_type;
  typedef BDom<float, Int> splainvector_type;
  typedef BDom<SmallVector<float, 3>, Int>   sbvector_type;
  matrix_type *pmat = NULL;
  smatrix_type *spmat = NULL;


  
  const int fcs=30;
  const int scs=30;
  typedef std::auto_ptr<PCGSolverBase<bvector_type> > solv_tp;
  solv_tp solv =
    Options::GetOptions().gcg? (solv_tp) new GPCGSolver<bvector_type> : (solv_tp) new PCGSolver<bvector_type>;
  PCGSolverBase<bvector_type>  &solver = *solv;
  solver.SetEps(eps);
  solver.SetSizes(N1, N2, Lines, N3R);
  solver.SetComm(comm);
  solver.SetMatrix(mat);
  ClearVector(x,1);
  //random initial guess
  PreconBase<bvector_type> *precon = NULL;
  PreconBase<bvector_type> *sprecon = NULL;
#ifdef _FLOAT_EXPAND
  typedef SinglePrecon<bvector_type, PreconBase<sbvector_type> > sprecon_type;
#endif
  {
    Int NumberOfLines = Lines+1;
    Int SL = StartLine-1;
    Int SO = -1;
    if (rank == 0)
    {
      SL++;
      SO++;
      NumberOfLines--;
    }
    if (rank +1 == size)
    {
      NumberOfLines--;
    }


    coefficientvec.Clear();
    coefficientvec.SetSize(N1, N2, NumberOfLines);
    coefficientvec.SetPars(
 			   0//Options::GetOptions().zero
 			   ,
			   0, Options::GetOptions().DM1,
			   0, Options::GetOptions().DM2,
			   Options::GetOptions().DM3*(double)SL/N3R,
			   Options::GetOptions().DM3*(double)(NumberOfLines+SL)/N3R);

    int off = FirstFull?1:0;
    if (rank==0)off=0;
    //    std::cout <<rank << " " <<NumberOfLines<<" "<<SL <<" " << DoubleLines<<"-------------------- "<< off <<std::endl;
    //    std::cout << DoubleLines <<" ------- " << NumberOfLines << std::endl;
    for(Int i = 0; i < N1; ++i)
      for(Int j = 0; j < N2; ++j)
	for(Int k  = 0; k<DoubleLines; ++k)
	{
	  double E, lambda, mu, ni;
	  //	  std::cout << i << " " << j << " " << k << std::endl;
	  lambda = mat->elmentw.GetAt(i/2, j/2, k).lambda;
	  mu =  mat->elmentw.GetAt(i/2, j/2, k).mu;
	  E = mu* ((3*lambda + 2*mu)/(lambda+mu));
	  ni = lambda/(2.*(lambda+mu));
	  //	  std::cout << E <<" " << ni << " " << lambda << " " << mu << std::endl;
	  coefficientvec(i,j,off+ k*2) = 
	  
	    coefficientvec(i,j,off+ k*2+1) = E;
	}

    int UpP = rank+1;
    int DownP = rank-1;
    if (DownP <0) DownP = MPI::PROC_NULL;
    if (UpP == size) UpP = MPI::PROC_NULL;

    MPI::Request rsd = MPI::REQUEST_NULL;
    MPI::Request rsu = rsd;
    int cnt = N1*N2;
    bool SendDown = (StartLine & 1) != 0;
    bool SendUp   = (((StartLine+Lines-1)&1) != 0);
    if ((rank+1)!=size)
      if(SendUp)
      {
	rsu = comm.Isend(&(coefficientvec(0,0,NumberOfLines-1)), cnt,
			 MPI::DOUBLE,
			 UpP, 0);
      }
      else
      {
	rsu = comm.Irecv(&(coefficientvec(0,0,NumberOfLines-1)), cnt,
			 MPI::DOUBLE,
			 UpP, 1);
      }
    

    if(rank)
      if(SendDown)
      {
	rsd = comm.Isend(&(coefficientvec(0,0,0)), cnt,
			 MPI::DOUBLE,
			 DownP, 1);
      }
      else
      {
	rsd = comm.Irecv(&(coefficientvec(0,0,0)), cnt,
			 MPI::DOUBLE,
			 DownP, 0);
      }
    
    rsd.Wait();
    rsu.Wait();
#ifdef _FLOAT_EXPAND
    
    if (Options::GetOptions().SinglePrecPrecon)
    {
      splainvector_type rhsp;
      rhsp.SetSize(N1, N2, Lines, N3R);
      rhsp.SetComm(comm);
      spmat = new smatrix_type;
      spmat->SetSize(N1, N2, Lines, N3R);
      spmat->SetComm(comm);
    
      ClearMatrix(*spmat);
      ClearVector(rhsp);
      DiscretizationType dt = distype;
//       if (distype==MP)dt = RP;
//       if (distype==MV)dt = RV;
      Discretizator<float> dis(dt);
      TimeCalc1("PreconConstruct");

      if((method==smPCG1)||(method==smPCG2)||(method==smPCG3)||(method==smPCG4)||(method==smPCG3PBC))
      {
	TimeCalc1("PreconConstruct");
	dis.GenerateMatrix(*spmat, rhsp, NumberOfLines, SL, SO, N3R, bc);
      }
      
    }
    else
#endif
    {
      plainvector_type rhsp;
      rhsp.SetSize(N1, N2, Lines, N3R);
      rhsp.SetComm(comm);
      pmat = new matrix_type;
      pmat->SetSize(N1, N2, Lines, N3R);
      pmat->SetComm(comm);
    
      ClearMatrix(*pmat);
      ClearVector(rhsp);
      DiscretizationType dt = distype;
//       if (distype==MP)dt = RP;
//       if (distype==MV)dt = RV;
      Discretizator<Double> dis(dt);
      if((method==smPCG1)||(method==smPCG2)||(method==smPCG3)||(method==smPCG4)||(method==smPCG3PBC))
      {
	TimeCalc1("PreconConstruct");
	dis.GenerateMatrix(*pmat, rhsp, NumberOfLines, SL, SO, N3R, bc);
      }
    }
    
  }
#ifdef _FLOAT_EXPAND
  if (Options::GetOptions().SinglePrecPrecon)
  {
    typedef SmallVector<float, 6> row_type;
    typedef BDom<srow_type, Int> matrix_type;
    typedef BDom<float, Int> plainvector_type;
    typedef sbvector_type bvector_type;
    PreconBase<bvector_type> *precon = NULL;
    switch(method)
    {
    case smCG:
      precon = new NoPrecon<bvector_type>();
      break;
    case smJacobi:
      precon = new JPrecon<bvector_type, BDom<SmallVector<Double, 3>,Int> >(mat);
      break;
      
    case smPCG1:
      {
	MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM1> *p
	  = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM1>;
	p->Create(*spmat);
	precon = p;
      }    
      break;
    case smPCG2:
      {
	MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2> *p
	  = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2>;
	p->Create(*spmat);
	precon = p;
      }    
      
      break;
    case smPCG3:
      {
	MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3> *p
	  = new MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3>;
	p->Create(*spmat);
	precon = p;
      }    
      
      break;
    case smPCG3PBC:
      {
	MIC0PreconPBC<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3PBC> *p
	  = new MIC0PreconPBC<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3PBC>;
	p->Create(*spmat);
	precon = p;
      }    
      
      break;
    case smPCG4:
      {
	MIC0PreconM4<matrix_type, bvector_type, plainvector_type, PreconDescriptorM4> *p
	  = new MIC0PreconM4<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM4>;
	p->Create(*spmat);
	//      std::cout << "m4 " << std::endl;
	precon = p;
      }
    default:;
    }
    sprecon = new sprecon_type(precon);
  }
  else
#endif
  {
#ifdef MY_HYPRE
    /// for hypreprecon discretizator
    Int NumberOfLines = Lines+1;
    Int SL = StartLine-1;
    Int SO = -1;
    if (rank == 0)
    {
      SL++;
      SO++;
      NumberOfLines--;
    }
    if (rank +1 == size)
    {
      NumberOfLines--;
    }
    DiscretizationType dt=distype;
//     if (distype==MP)dt = RP;
//     if (distype==MV)dt = RV;

    Discretizator<double> dis(dt);
    HypreDDPrecon<bvector_type>* hp = NULL;      
    ///end for it
#endif
    switch(method)
    {
#ifdef MY_HYPRE
    case smBoomerDD:
      {
	TimeCalc1("PreconConstruct");

	hp = new HypreDDPrecon<bvector_type>;
 	hp->Init(comm, N1, N2, N3R);
 	dis.GenerateMatrix(hp->mat, hp->hdst, NumberOfLines, SL, SO, N3R, bc);
 	hp->Init2();
	precon = hp;
      }
      break;
#endif
    case smCG:
      precon = new NoPrecon<bvector_type>();
      break;
    case smJacobi:
      precon = new JPrecon<bvector_type>(mat);
      break;
      
    case smPCG1:
      {
	MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM1> *p
	  = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM1>;
	p->Create(*pmat);
	precon = p;
      }    
      break;
    case smPCG2:
      {
	MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2> *p
	  = new MIC0Precon<matrix_type, bvector_type, plainvector_type, PreconDescriptorM2>;
	p->Create(*pmat);
	precon = p;
      }    
      
      break;
    case smPCG3:
      {
	MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3> *p
	  = new MIC0Precon<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3>;
	p->Create(*pmat);
	precon = p;
      }    
      
      break;
    case smPCG3PBC:
      {
	MIC0PreconPBC<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3PBC> *p
	  = new MIC0PreconPBC<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM3PBC>;
	p->Create(*pmat);
	precon = p;
      }    
      
      break;
    case smPCG4:
      {
	MIC0PreconM4<matrix_type, bvector_type, plainvector_type, PreconDescriptorM4> *p
	  = new MIC0PreconM4<matrix_type, bvector_type, plainvector_type,  PreconDescriptorM4>;
	p->Create(*pmat);
	//      std::cout << "m4 " << std::endl;
	precon = p;
      }    
    default:;
    }
  }



  if (sprecon)
  {
       precon = sprecon;
  }
  solver.SetPrecon(precon);

  coefficientvec.Clear();


  if (rank == 0)
  {
     int proc = 1;
     #ifdef _OPENMP
     proc =    omp_get_max_threads();
     #endif
    cout << "Starting  on " 
     << size << " processors, " << proc <<" omp processors. Eps is "<< eps<< endl;
    
    cout << setw(fcs) << left << "Problem size "
	 << N1 << "x" << N2 << "x" << N3R << " " << Myf<Double>::ToString() << " "
	 << algname + solver.GetName() <<endl;
    cout << setw(fcs) << left  << "Running on "  
	 << setw(scs) <<size  << "procesors." << endl;
  }
  algname += solver.GetName();
  int iterations;
  {
    TimeCalc1("!Solution time");
    iterations  = solver.Solve(x, rhs);
  }
  if (rank == 0)
  {
    cout << setw(fcs) << "Iterations" 
	 << setw(scs) << iterations << endl;
  }
  delete precon;
  delete pmat;
  return iterations;
}







#endif

