#ifndef __smallops_h__
#define __smallops_h__
#include "small.h"
#include <cmath>
#include <cassert>

template <class T, int N> inline void LU(SmallMatrix<T, N, N> &a, SmallVector<int, N> & index, T &d)
{
  SmallVector<T, N> vv;
    d = 1.0;
    for (int i = N; i--; ) {
        T max = T();
        for (int j = N; j--; )
            max = std::max(std::fabs(a[i][j]), max);
        vv[i] = 1./max;
    }
    for (int j = 0; j < N; ++j) {
        for (int i = 0; i < j; ++i) {
            T s = a[i][j];
            for (int k = 0; k < i; ++k)
                s -= a[i][k]*a[k][j];
            a[i][j] = s;
        }
        T max = T();
        int imax = 0;
        for (int i = j; i < N; ++i) {
            T s = a[i][j];
            for (int k = 0; k < j; ++k)
                s -= a[i][k]*a[k][j];
            a[i][j] = s;
            T tmp;
            if ((tmp = vv[i]*std::fabs(s)) >= max) {
                max = tmp;
                imax = i;
            }
        }
        if (j != imax) {
            for (int k = 0; k < N; ++k) {
                T tmp = a[imax][k];
                a[imax][k] = a[j][k];
                a[j][k] = tmp;
            }
            d = -d;
            vv[imax] = vv[j];
        }
        index[j] = imax;
        if (j != N - 1) {
            T tmp = 1./a[j][j];
            for (int i = j + 1; i < N; ++i)
                a[i][j] *= tmp;
        }
    }
}

// LU back substitution
template <class T, class V, int N> 
  inline void LUsolve(SmallMatrix<T, N, N> &a, SmallVector<int, N> & index, SmallVector<V, N> & b)
{
    int k = -1;
    for (int i = 0; i < N; ++i) {
        int pos = index[i];
        T s = b[pos];
        b[pos] = b[i];
        if (k >= 0) {
            for (int j = k; j < i; ++j)
                s -= a[i][j]*b[j];
        } else if (s) {
            k = i;
        }
        b[i] = s;
    }
    for (int i = N; i--;) {
        T s = b[i];
        for (int j = i + 1; j < N; ++j)
            s -= a[i][j]*b[j];
        b[i] = s/a[i][i];
    }
}

template <class T, int N> inline T LUinverse(SmallMatrix<T, N, N> &a)
{
    T det;
    SmallMatrix<T, N, N> m;
    for (int i = N; i--; )
        for (int j = N; j--; )
            m[i][j] = a[i][j];
    SmallVector<int, N> index;
    LU(m, index, det);
    for (int i = N; i--; )
        det *= m[i][i];
    for (int j = 0; j < N; ++j) {
      SmallVector<T, N> c;c = T();
        c[j] = 1.;
        LUsolve(m, index, c);
        for (int i = N; i--; )
            a[i][j] = c[i];
    }
    return det;
}

// determinant
template <class T, int N> 
inline T LUdet(const SmallMatrix<T, N, N> &a, const SmallVector<int, N>, T  det)
{
    for (int i = N; i--; )
        det *= a[i][i];
    return det;
}



// matrix-matrix multiplication
template <class T, class U, class V, int N1, int N2, int N3> 
  inline void MMmul(SmallMatrix<T, N1, N3> &r, SmallMatrix<U, N1, N2> const &a, SmallMatrix<V, N2, N3> const &b)
{
  r = T();
  for (int i = N1; i--; ) {
    for (int k = N2; k--; ) {
      for (int j = N3; j--; ) {
	r[i][j] += a[i][k]*b[k][j];
      }
    }
  }
}


template <class T, int N>
inline SmallMatrix<T, N, N> operator *(SmallMatrix<T, N, N> const &a,
				       SmallMatrix<T, N, N> const &b)
{
  SmallMatrix<T, N, N> res;
  MMmul(res, a, b);
  return res;
}




template <class T, class U, class V, int N1, int N2, int N3> 
  inline void MTmul(SmallMatrix<T, N1, N3> &r, SmallMatrix<U, N1, N2> const &a, SmallMatrix<V, N3, N2> const &b)
{
  r = T();
  for (int i = N1; i--; ) {
    for (int k = N2; k--; ) {
      for (int j = N3; j--; ) {
	r[i][j] += a[i][k]*b[j][k];
      }
    }
  }
}



// transpose matrix-matrix multiplication
template <class T, class U, class V, int N1, int N2, int N3> 
inline void TMmul(SmallMatrix<T, N1, N3> &r, SmallMatrix<U, N2, N1> const&a, SmallMatrix<V, N2, N3> const&b)
{
  r = T();
  for (int i = N1; i--; ) {
    for (int k = N2; k--; ) {
      for (int j = N3; j--; ) {
	r[i][j] += a[k][i]*b[k][j];
      }
    }
  }
}

template <class T, class U, int N> 
  inline void transform(SmallMatrix<T, N, N> &a, SmallMatrix<U, N, N> const&h)
{
  SmallMatrix<T, N, N> t;
  TMmul(t, h, a);
  MMmul(a, t, h);
}
template <class T, class U, int N> 
  inline void left_transform(SmallMatrix<T, N, N> &a, SmallMatrix<U, N, N> &h)
{
  SmallMatrix<T, N, N> t;
  TMmul(t, a, h);
  a = t;
}

template<class T, int N>
inline T VecDot(SmallVector<T, N> const &v1, SmallVector<T, N> const&v2)
{
  T res = T();
  for (int i = N; i--;)
    res+= v1[i]*v2[i];
  return res;
}



template<class T>
inline SmallVector<T, 3> const VecCross(SmallVector<T, 3> const &v1,
				  SmallVector<T, 3> const &v2)
{
  SmallVector<T, 3> r;
  r[0] = v1[1]*v2[2] - v1[2]*v2[1];
  r[1] = v1[2]*v2[0] - v1[0]*v2[2];
  r[2] = v1[0]*v2[1] - v1[1]*v2[0];
  return r;
}


template <class T, int N1, int N2>
  inline void MatVec(SmallMatrix<T, N1, N2> const &a, SmallVector<T, N2> const &h, SmallVector<T, N1> &r)
{
  for (int i = N1; i--;)
  {
    r[i] = VecDot(a[i], h);
    /* r[i] = T(); */
    /* for(int j = 0; j < N2; ++j) */
    /*   r[i]+=a[i][j]*h[j]; */
  }
}

template <class T, int N1, int N2>
  inline void VecMat(SmallMatrix<T, N1, N2> const &a, SmallVector<T, N2> const &h, SmallVector<T, N1> &r)
{
  for (int i = 0; i < N1; ++i)
  {
    r[i] = T();
    for(int j = 0; j < N2; ++j)
      r[i]+=a[j][i]*h[j];
  }
}
template <class T, int N>
inline SmallVector<T, N> operator *(SmallMatrix<T, N, N> const &a,
				    SmallVector<T, N> const &b)
{
  SmallVector<T, N> res;
  MatVec(a, b, res);
  return res;
}



template<class T>
void VecExtend(SmallVector<T, 4> &r, const SmallVector<T, 4>&s)
{
  r = s;
}
template<class T>
void VecExtend(SmallVector<T, 4> &r, const SmallVector<T, 3>&s)
{
  for (int i = 0; i < 3; ++i)
    r[i] = s[i];
  r[3] = 1.;
}
template<class T>
void VecShrink(SmallVector<T, 3> &r, const SmallVector<T, 4>&s)
{
  for (int i = 0; i < 3; ++i)
    r[i] = s[i]/s[3];
}
template<class T>
void VecShrink(SmallVector<T, 3> &r, const SmallVector<T, 3>&s)
{
  r = s;
}
template<class T>
inline SmallVector<T, 3> operator *(SmallMatrix<T, 4, 4> const &a,
				    SmallVector<T, 3> const &b)
{
  SmallVector<T, 4> tmp, res4;
  SmallVector<T, 3>  res3;
  VecExtend(tmp, b);
  MatVec(a, tmp, res4);
  VecShrink(res3, res4);
  return res3;
}

template<class T, int N>
T  Norm2(SmallVector<T, N> const &r)
{
  T norm = T();
  for (int i = 0; i < 3; ++i)
    norm += r[i]*r[i];
  norm = sqrt(norm);
  return norm;
}
template<class T>
void Normalize(SmallVector<T, 3> &r)
{
  T size = T();
  for (int i = 0; i < 3; ++i)
    size += r[i]*r[i];
  if (size >1e-16)
    r/= sqrt(size);
  else
    r = 0;
}

template<class T>
void Normalize(SmallVector<T, 4> &r)
{
  T size = T();
  for (int i = 0; i < 3; ++i)
    size += r[i]*r[i];
  //  r/=sqrt(size);
  r[3] = sqrt(size);
}


template<class T, int N>
void TranslationMatrix(SmallMatrix<T, 4, 4> &a, const SmallVector<T, N>&v)
{
  assert((N==3) || (N==4));
  a = T();
  SmallVector<T, 3> V;
  VecShrink(V, v);
  for (int i = 0; i < 4; ++i)
    a[i][i] = 1.;
  for (int i = 0; i < 3; ++i)
    a[i][3] = V[i];  
}

// translation from point A, to point B;
template<class T, int N>
void TranslationMatrix(SmallMatrix<T, 4, 4> &a, const SmallVector<T, N>&A,
		       const SmallVector<T, N>&B)
{
  assert((N==3) || (N==4));
  a = T();
  SmallVector<T, 3> NA, NB;
  VecShrink(NA, A);
  VecShrink(NB, B);
  NB-=NA;
  TranslationMatrix(a, NB);
}



template<class T>
void ScaleMatrix(SmallMatrix<T, 4, 4> &a, T factor)
{
  a = T();
  for (int i = 0; i < 3; ++i)
    a[i][i] = factor;
  a[3][3] = 1.;
}

template<class T>
void IdentityMatrix(SmallMatrix<T, 4, 4> &a)
{
  ScaleMatrix(a, 1);
}

// rotation  from v1 to v2;
template<class T, int N>
void VecRotationMatrix(SmallMatrix<T, 4, 4> &a, const SmallVector<T, N>&v1, const SmallVector<T, N>&v2,   const SmallVector<T, N>&r1, const SmallVector<T, N> &r2)
{
  SmallVector<T, 3> cross3;
  SmallVector<T, 3> V1, V2, C1, C2, R1, R2;
  SmallMatrix<T, 3, 3> L;
  SmallMatrix<T, 3, 3> R;
  SmallMatrix<T, 3, 3> res;
  VecShrink(V1, v1);
  VecShrink(V2, v2);
  VecShrink(R1, r1);
  VecShrink(R2, r2);
  Normalize(V1);
  Normalize(V2);
  Normalize(R1);
  Normalize(R2);
  
  C1 = VecCross(V1, R1);
  C2 = VecCross(V2, R2);
  // if (Norm2(cross3)<1e-16)
  // {
  //   //    std::cout << "WWWWWWWWWWWWWWWWarning colinear vectors!" << std::endl;
  //   T scale = Norm2(V2)/Norm2(V1);
  //   ScaleMatrix(a, scale);
  //   return;
  // }
  
  // we have AL = R, where L = [V1 R1 C1] and R = [V2 R2 C2]
  for (int i = 0; i < 3; ++i)
  {
    L[i][0] = V1[i];
    R[i][0] = V2[i];

    L[i][1] = R1[i];
    R[i][1] = R2[i];

    L[i][2] = C1[i];
    R[i][2] = C2[i];    
  }
  LUinverse(L);

  MMmul(res, R, L);

  a=T();
  for (int i = 0; i < 3; ++i)
    for (int j = 0; j < 3; ++j)
      a[i][j] = res[i][j];

  a[3][3] = 1.;

  // std::cout << "V1 = " << V1 << std::endl;
  // std::cout << "V2 = " << V2 << std::endl;
  // std::cout << "R1 = " << R1 << std::endl;
  // std::cout << "R2 = " << R2 << std::endl;
  // std::cout << "C1 = " << C1 << std::endl;
  // std::cout << "C2 = " << C2 << std::endl;
  // C2 = a * C1;
  // V2 = a * V1;
  // R2 = a * R1;
  // std::cout << "NewV2 = " << V2 << std::endl;
  // std::cout << "NewR2 = " << R2 << std::endl;
  // std::cout << "NewC2 = " << C2 << std::endl;
  
}


//Makes the vector from point A1 to point A2
template<class T, int N>
void Vector(SmallVector<T, 3> &r,
	    const SmallVector<T, N> &A1,
	    const SmallVector<T, N> &A2)
{
  SmallVector<T, 3> r1;
  VecShrink(r1, A1);
  VecShrink(r, A2);
  r-=r1;
}

template<class T>
T VecLength(const SmallVector<T, 3> &v)
{
  return sqrt(VecDot(v, v));
}


// transformation maps A1->B1 and A2 -> b2;
template<class T, int N>
void TransformationMatrix(SmallMatrix<T, 4, 4> &a,
			  const SmallVector<T, N> &A1,
			  const SmallVector<T, N> &A2,
			  const SmallVector<T, N> &B1,
			  const SmallVector<T, N> &B2,
			  const SmallVector<T, N> &or1,
			  const SmallVector<T, N> &or2,
			  T ScaleFactor=T())
{
  SmallVector<T, 3> zero, V1, V2;
  zero = T();
  SmallMatrix<T, 4, 4> tran1, tran2, rot, scale;
  Vector(V1, A1, A2);
  Vector(V2, B1, B2);
  TranslationMatrix(tran1, A1, zero);
  VecRotationMatrix(rot, V1, V2, or1, or2);
  if(!ScaleFactor)
  {
    ScaleFactor = VecLength(V2)/VecLength(V1);
  }
  ScaleMatrix(scale, ScaleFactor);
  TranslationMatrix(tran2, zero, B1);
  a = tran2 * scale * rot * tran1;//
}


template<class T, int N, int M>
void Transform(SmallMatrix<T, 4, 4> const &mat, SmallVector<T, N> const & from,
	       SmallVector<T, M> &to)
{
  SmallVector<T, 4> From, To;
  VecExtend(From, from);
  MatVec(mat, From, To);
  VecShrink(to,To);
}

#endif
