#ifndef __solver_h__
#define __solver_h__
#include "basic.h"
#include "mpi.h"
#include <string>
#include "useful.h"
#include <vector>
#include <valarray>
#include <cassert>
#include "projection.h"

template<class Int>
struct CircularIndex
{
  Int a;
  Int w; //how much to wrap around
  explicit  CircularIndex(Int b):a(0), w(b){assert(w>0);} 
  CircularIndex(CircularIndex const &b):a(b.a), w(b.w){}
  CircularIndex& operator = (CircularIndex const & b)
  {
    a = b.a;
    w = b.w;
    return *this;
  }
  CircularIndex& operator = (Int i)
  {
    a = i;
    Normalize();
    return *this;
  }
  void Normalize()
  {
    a = a % w;
    if (a<0)
      a += w;
  }
  operator Int(){return a;}
  //friend operators but.... next time;
};







template<class Vector>
class PCGSolverBase;

EXPORT
template<class Vector>
class MatrixBase
{
 public:
  virtual void Multiply(Vector &res, Vector &src)const = 0;
  virtual ~MatrixBase(){};
  virtual std::string GetName()const = 0;
};
EXPORT
template<class Vector>
class PreconBase
{
 public:
  typedef Vector vector_type;
  virtual void PreconSolve(Vector &res, Vector &src)const = 0;
  virtual ~PreconBase(){};
  virtual std::string GetName()const = 0;
};

EXPORT
template<class Vector>
class PCGSolverBase
{
public:
  typedef Vector vector_type;
  typedef typename Myf<typename Vector::base_type>::base_type Double;
  typedef typename Myf<typename Vector::base_type>::type element_type;
  typedef typename Vector::Int Int;
protected:
  MatrixBase<Vector> *mat;
  PreconBase<Vector> *precon;
  Double eps;

 public:


  void SetMatrix(MatrixBase<Vector> *m){mat = m;}
  void SetPrecon(PreconBase<Vector> *p){precon = p;}
  PCGSolverBase();
  void SetEps(Double eps_);

  virtual void SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_)=0;
  virtual void SetComm(MPI::Intracomm &comm_)=0;
  virtual unsigned Solve(Vector &X, Vector  &RHS)=0;
  
  //operations Better to be global functions associated with the vectors themselves
  virtual std::string GetName()const=0;
  virtual ~PCGSolverBase(){}
  
};



EXPORT
template<class Vector>
class PCGSolver:public PCGSolverBase<Vector>
{
public:
  typedef Vector vector_type;
  typedef typename Myf<typename Vector::base_type>::base_type Double;
  typedef typename Myf<typename Vector::base_type>::type element_type;
  typedef typename Vector::Int Int;
protected:
  using PCGSolverBase<Vector>::mat;
  using PCGSolverBase<Vector>::precon;
  using PCGSolverBase<Vector>::eps;
  Projection<Vector> proj;

  vector_type Hd, D, G;
public:
  virtual void SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_);
  virtual void SetComm(MPI::Intracomm &comm_);

  //operations Better to be global functions associated with the vectors themselves
  virtual unsigned Solve(Vector &X, Vector  &RHS);
  virtual std::string GetName()const
  {
    return "PCG" + mat->GetName() + precon->GetName();
  }
};

EXPORT
template<class Vector>
class GPCGSolver:public PCGSolverBase<Vector>
{
private:
  int m;
protected:
  Projection<Vector> proj;

public:
  using PCGSolverBase<Vector>::mat;
  using PCGSolverBase<Vector>::eps;
  using PCGSolverBase<Vector>::precon;
  
  typedef Vector vector_type;
  typedef typename Myf<typename Vector::base_type>::base_type Double;
  typedef typename Myf<typename Vector::base_type>::type element_type;
  typedef typename Vector::Int Int;

  GPCGSolver();
  
  vector_type Hd, G, h;
  vector_type * Ds;
  std::vector<Double> taus;
  std::vector<Double> sigmas;
  virtual void SetSizes(Int N1_, Int N2_, Int N3_, Int N3R_);
  virtual void SetComm(MPI::Intracomm &comm_);

  //operations Better to be global functions associated with the vectors themselves
  virtual unsigned Solve(Vector &X, Vector  &RHS);
  virtual std::string GetName()const
  {
    std::ostringstream os;
    os <<  "GPCG" << m <<  mat->GetName() << precon->GetName();
      return os.str();
  }
  ~GPCGSolver(){delete[] Ds;}
  
};


#endif
#ifdef INCLUSION
#include "solver.hpp"
#endif
